import os
from typing import List, Union, Dict
import json
import logging
from asyncio import AbstractEventLoop

from joule.api.node.tcp_node import TcpNode
from joule.api.node.unix_node import UnixNode
from joule.api.node.node_config import NodeConfig
from joule import errors

from joule.api.node.base_node import BaseNode


def get_node(name: str = "") -> BaseNode:
    try:
        configs = _get_node_configs()
        if name == "":
            config = _get_default_node(configs)
        else:
            config = configs[name]
    except ValueError as e:
        raise errors.ApiError(str(e))
    except KeyError as e:
        raise errors.ApiError("Node [%s] is not available, add it with [joule master add]" % str(e))
    return TcpNode(config.name, config.url, config.key, _get_cafile())


def get_nodes() -> List[BaseNode]:
    nodes = []
    cafile = _get_cafile()
    for config in _get_node_configs().values():
        my_node = TcpNode(config.name,
                          config.url,
                          config.key,
                          cafile=cafile
                          )
        nodes.append(my_node)
    return nodes


def save_node(node: TcpNode) -> None:
    configs = _get_node_configs()
    configs[node.name] = node.to_config()
    _write_node_configs(configs)


def delete_node(node: Union[str, BaseNode]) -> None:
    configs = _get_node_configs()
    if type(node) is str:
        name = node
    else:
        name = node.name
    if name not in configs:
        raise errors.ApiError("Node [%s] does not exist" % name)
    del configs[name]
    _write_node_configs(configs)
    # if this is the default node, pick a new one
    try:
        default_config = _get_default_node(configs)
        if default_config.name == name:
            set_default_node("")
    except ValueError:
        # this was the last node, no other nodes available
        pass


def set_default_node(node: Union[str, BaseNode]) -> None:
    if type(node) is str:
        name = node
    else:
        name = node.name

    config_dir = _user_config_dir()
    default_node_path = os.path.join(config_dir, "default_node.txt")

    node_configs = _get_node_configs()
    if len(name) > 0 and name not in node_configs.keys():
        raise ValueError("Invalid node name, view nodes with [joule node list]")
    if name == "":
        if len(node_configs) == 0:
            return  # nothing to do, no nodes available
        # pick a new default node
        with open(default_node_path, 'w') as f:
            first_name = list(node_configs.keys())[0]
            f.write(first_name + "\n")
            return

    with open(default_node_path, 'w') as f:
        f.write(name + "\n")


def create_tcp_node(url: str, key: str, name: str = "node",
                    loop: AbstractEventLoop = None) -> TcpNode:
    return TcpNode(name, url, key, _get_cafile(), loop)


def create_unix_node(path: str, name: str = "node",
                     loop: AbstractEventLoop = None) -> BaseNode:
    return UnixNode(name, path, loop)


def _get_cafile():
    config_dir = _user_config_dir()
    cafile = os.path.join(config_dir, "ca.crt")
    if os.path.isfile(cafile):
        return cafile
    else:
        return ""


def _get_node_configs() -> Dict[str, NodeConfig]:
    node_configs = {}
    config_dir = _user_config_dir()
    nodes_path = os.path.join(config_dir, "nodes.json")

    # if nodes.json does not exist try to create it

    try:
        with open(nodes_path, 'r')as f:
            nodes = json.load(f)
        for node in nodes:
            node_configs[node['name']] = NodeConfig(node['name'],
                                                    node['url'],
                                                    node['key'])
    except FileNotFoundError:
        node_configs = {}
    except json.decoder.JSONDecodeError:
        raise ValueError("Cannot parse [%s], fix syntax or remove it" % nodes_path)

    return node_configs


def _get_default_node(configs: Dict[str, NodeConfig]) -> NodeConfig:
    # if default_node.txt is empty or does not exist set it
    # to the first entry in nodes
    config_dir = _user_config_dir()
    default_node_path = os.path.join(config_dir, "default_node.txt")
    # Case 1: No nodes in nodes.json
    if len(configs) == 0:
        # no nodes, remove the default
        with open(default_node_path, 'w') as f:
            f.write("")
        os.chmod(default_node_path, 0o600)
        raise ValueError("No nodes available, use [joule admin authorize] or [joule node add]")
    # Case 2: default_node.txt does not exist
    if not os.path.isfile(default_node_path):
        with open(default_node_path, 'w') as f:
            first_name = list(configs.keys())[0]
            f.write(first_name + "\n")
        os.chmod(default_node_path, 0o600)
        return configs[first_name]

    # Case 3: Default is not in the nodes.json file
    with open(default_node_path, 'r') as f:
        name = f.readline().strip()
    if name not in configs:
        # change the default node to a valid choice
        with open(default_node_path, 'w') as f:
            first_name = list(configs.keys())[0]
            f.write(first_name + "\n")
        os.chmod(default_node_path, 0o600)
        return configs[first_name]

    # OK, return the default node
    return configs[name]


def _write_node_configs(node_configs: Dict[str, NodeConfig]):
    config_dir = _user_config_dir()
    nodes_path = os.path.join(config_dir, "nodes.json")
    default_node_path = os.path.join(config_dir, "default_node.txt")

    # write out the key into nodes.json
    with open(nodes_path, "w") as f:
        json_val = [n.to_json() for n in node_configs.values()]
        json.dump(json_val, f, indent=2)
    os.chmod(nodes_path, 0o600)

    # if the default file is missing or empty, add a node as the default
    if not os.path.isfile(default_node_path):
        with open(default_node_path, 'w') as f:
            f.write("")
    with open(default_node_path, 'r') as f:
        name = f.readline()
    if name == "":
        with open(default_node_path, 'w') as f:
            first_name = list(node_configs.keys())[0]
            f.write(first_name + "\n")
    os.chmod(default_node_path, 0o600)
    _fix_config_ownership()

# save the database entry now that everything is written out


def _fix_config_ownership():
    # fix permissions if this was run by root
    if 'SUDO_USER' not in os.environ:
        return

    uid = int(os.environ["SUDO_UID"])
    gid = int(os.environ["SUDO_GID"])

    config_dir = _user_config_dir()
    os.chown(config_dir, uid, gid)
    nodes_path = os.path.join(config_dir, "nodes.json")
    os.chown(nodes_path, uid, gid)
    default_node_path = os.path.join(config_dir, "default_node.txt")
    os.chown(default_node_path, uid, gid)


def _user_config_dir() -> str:
    # return path to user config directory
    # create it if it doesn't exist
    if "JOULE_USER_CONFIG_DIR" in os.environ:
        config_dir = os.environ["JOULE_USER_CONFIG_DIR"]
    else:
        config_dir = os.path.join(os.environ["HOME"], ".joule")
    if not os.path.isdir(config_dir):
        os.mkdir(config_dir, mode=0o700)
    return config_dir



import click
from joule.cli.config import pass_config
from joule import api
from joule import errors


@click.command(name="delete")
@click.argument("name")
@pass_config
def node_delete(config, name):
    try:
        api.delete_node(name)
    except errors.ApiError as e:
        raise click.ClickException(str(e))
    click.echo("Removed [%s] from nodes" % name)



import aiodns
import aiohttp
from aiohttp import web
from sqlalchemy.orm import Session

from joule.models import Follower
from joule.api import TcpNode

async def index(request: web.Request):
    db: Session = request.app["db"]
    followers = db.query(Follower)
    return web.json_response([f.to_json() for f in followers])


# Note: this is covered with e2e tests
async def add(request: web.Request):  # pragma: no cover
    """
    Called by a Joule node to allow *this* node to control it
    Parameters:
        key(str): the API key to use
        base_uri(str): (optional) specify the URL of the follower
        port(int): specify the port Joule is using on the follower

    """
    db: Session = request.app["db"]
    cafile = request.app["cafile"]
    if request.content_type != 'application/json':
        return web.Response(text='content-type must be application/json', status=400)
    body = await request.json()

    try:
        key = body["key"]
        port = body["port"]
        scheme = body["scheme"]
        base_uri = body["base_uri"]
        if "name_is_host" in body:
            location = scheme+"://" + body["name"] + ":" + str(port) + base_uri
        else:
            location = scheme+"://" + request.remote + ":" + str(port) + base_uri
        try:
            node = TcpNode("new_follower", location, key, cafile)
            info = await node.info()
            await node.close()
            follower_name = info.name
        except aiohttp.ClientError:
            return web.Response(text="no route to node", status=400)
        except ValueError as e:
            return web.Response(text=str(e), status=400)
        follower = Follower(key=key, location=location,
                            name=follower_name)
        # update the follower if this is a repeat
        db.query(Follower).filter(Follower.location == follower.location).delete()
        db.add(follower)
        db.commit()
    except KeyError as e:
        return web.Response(text="Invalid request, missing [%s]" % str(e), status=400)
    except ValueError as e:
        return web.Response(text="Invalid request, bad argument format", status=400)
    return web.json_response(data={'name': request.app["name"]})


async def delete(request: web.Request):
    """
    Remove the specified node from the list of followers
    """
    db: Session = request.app["db"]
    try:
        name = request.query["name"]
        db.query(Follower).filter(name == name).delete()
        db.commit()
    except KeyError:
        return web.Response(text="specify name to remove", status=400)
    return web.Response(text="OK")



import click
import asyncio
import psutil
import os
import configparser
import sqlalchemy

from joule.cli.config import pass_config

WORKING_DIRECTORY = '/tmp/joule'


def abort_if_false(ctx, param, value):
    if not value:
        ctx.abort()


@click.command(name="erase")
@click.option("-c", "--config", help="main configuration file", default="/etc/joule/main.conf")
@click.option("-l", "--links", is_flag=True, help="delete masters and followers as well as data")
@click.option('--yes', is_flag=True, callback=abort_if_false,
              expose_value=False,
              prompt='Are you sure you want to wipe the local node?')
def admin_erase(config, links):
    from joule.services import load_config
    from joule.errors import ConfigurationError
    # make sure joule is not running
    pid_file = os.path.join(WORKING_DIRECTORY, 'pid')
    if os.path.exists(pid_file):
        with open(pid_file, 'r') as f:
            pid = int(f.readline())
            if psutil.pid_exists(pid):
                raise click.ClickException("stop joule service before running this command")

    # load the config file
    if os.path.isfile(config) is False:
        raise click.ClickException("Invalid configuration: cannot load file [%s]" % config)
    parser = configparser.ConfigParser()
    try:
        with open(config, 'r') as f:
            parser.read_file(f)
    except PermissionError:
        raise click.ClickException("insufficient permissions, run with [sudo]")
    except FileNotFoundError:
        raise click.ClickException("Joule config [%s] not found, specify with --config")

    try:
        joule_config = load_config.run(custom_values=parser)
    except ConfigurationError as e:
        raise click.ClickException("Invalid configuration: %s" % e)

    loop = asyncio.get_event_loop()
    loop.run_until_complete(run(joule_config,links))
    loop.close()


async def run(config, delete_links):
    from sqlalchemy import create_engine
    from sqlalchemy.orm import Session

    from joule.models import (Stream, Element,
                              Master, Follower, Annotation,
                              Folder, Base, TimescaleStore)
    from joule.errors import DataError

    # erase metadata
    engine = create_engine(config.database)
    with engine.connect() as conn:
        conn.execute('CREATE SCHEMA IF NOT EXISTS data')
        conn.execute('CREATE SCHEMA IF NOT EXISTS metadata')

    Base.metadata.create_all(engine)
    db = Session(bind=engine)
    db.query(Element).delete()
    db.query(Annotation).delete()
    db.query(Stream).delete()
    db.query(Folder).delete()
    if delete_links:
        db.query(Master).delete()
        db.query(Follower).delete()
    db.commit()

    if config.nilmdb_url is not None:
        click.echo("Not erasing NilmDB data")
        return

    # erase data
    data_store = TimescaleStore(config.database,
                                config.insert_period,
                                config.cleanup_period,
                                asyncio.get_event_loop())
    try:
        await data_store.initialize([])
        await data_store.destroy_all()
    except DataError as e:
        click.echo("Error erasing database")
        raise click.ClickException(str(e))



import click
import asyncio

from joule import errors
from joule.cli.config import pass_config


@click.command(name="delete")
@click.argument("type", type=click.Choice(['user', 'joule', 'lumen']))
@click.argument("name")
@pass_config
def cli_delete(config, type, name):
    loop = asyncio.get_event_loop()
    try:
        loop.run_until_complete(
            _run(config.node, type, name))
    except errors.ApiError as e:
        raise click.ClickException(str(e)) from e
    finally:
        loop.run_until_complete(
            config.close_node())
        loop.close()


async def _run(node, master_type, name):
    await node.master_delete(master_type, name)
    click.echo("Access to node [%s] revoked for %s [%s]" % (node.name, master_type, name))



from sqlalchemy.orm import Session
from typing import Dict ,List
import logging
import re

from joule.models import Stream
from joule.errors import ConfigurationError
from joule.models.folder import find as find_folder
from joule.models.stream import from_config as stream_from_config
from joule.services.helpers import (Configurations,
                                    load_configs)

logger = logging.getLogger('joule')

# types
Streams = Dict[str, List[Stream]]


def run(path: str, db: Session):
    configs = load_configs(path)
    configured_streams = _parse_configs(configs)
    for path, streams in configured_streams.items():
        for stream in streams:
            _save_stream(stream, path, db)
    return [stream for _, stream in configured_streams.items()]


def _parse_configs(configs: Configurations) -> Streams:
    streams: Streams = {}
    for file_path, data in configs.items():
        try:
            s = stream_from_config(data)
            s.is_configured = True
            stream_path = _validate_path(data['Main']['path'])
            if stream_path in streams:
                streams[stream_path].append(s)
            else:
                streams[stream_path] = [s]
        except KeyError as e:
            logger.error("Invalid stream [%s]: [Main] missing %s" %
                         (file_path, e.args[0]))
        except ConfigurationError as e:
            logger.error("Invalid stream [%s]: %s" % (file_path, e))
    return streams


def _validate_path(path: str) -> str:
    #
    if path != '/' and re.fullmatch(r'^(/[\w -]+)+$', path) is None:
        raise ConfigurationError(
            "invalid path, use format: /dir/subdir/../file. "
            "valid characters: [0-9,A-Z,a-z,_,-, ]")

    return path


def _save_stream(new_stream: Stream, path: str, db: Session) -> None:
    my_folder = find_folder(path, db, create=True)
    cur_stream: Stream = db.query(Stream) \
        .filter_by(folder=my_folder, name=new_stream.name) \
        .one_or_none()
    if cur_stream is not None:
        if cur_stream.layout != new_stream.layout:
            logger.error("Invalid layout %s for [%s], existing stream has layout %s" % (
                new_stream.layout, cur_stream.name, cur_stream.layout))
        else:
            cur_stream.merge_configs(new_stream)
            db.add(cur_stream)
            db.expunge(new_stream)
    else:
        my_folder.streams.append(new_stream)



from typing import Tuple, Dict, Optional, List
import asyncio
import click
import json
import logging

from joule.models import pipes
from joule.api import BaseNode, stream
from joule import errors
from joule.services.parse_pipe_config import parse_pipe_config, parse_inline_config
from joule.utilities import timestamp_to_human

Pipes = Dict[str, pipes.Pipe]
Loop = asyncio.AbstractEventLoop
log = logging.getLogger('joule')


async def build_fd_pipes(pipe_args: str, node: BaseNode) -> Tuple[Pipes, Pipes]:
    try:
        pipe_json = json.loads(json.loads(pipe_args))
        # if debugging, pycharm escapes the outer JSON
        # pipe_json = json.loads(pipe_args.encode('utf-8').decode('unicode_escape'))
        dest_args = pipe_json['outputs']
        src_args = pipe_json['inputs']
    except (KeyError, json.JSONDecodeError):
        raise errors.ConfigurationError("invalid pipes argument: [%s]" % pipe_args)
    pipes_out = {}
    pipes_in = {}
    for name, arg in dest_args.items():
        wf = pipes.writer_factory(arg['fd'], node.loop)
        dest_stream = None
        if arg['id'] is not None:  # used in testing when no API is available
            dest_stream = await node.stream_get(arg['id'])
        pipes_out[name] = pipes.OutputPipe(stream=dest_stream,
                                           layout=arg['layout'],
                                           writer_factory=wf)

    for name, arg in src_args.items():
        rf = pipes.reader_factory(arg['fd'], node.loop)
        src_stream = None
        if arg['id'] is not None:  # used in testing when no API is available
            src_stream = await node.stream_get(arg['id'])
        pipes_in[name] = pipes.InputPipe(stream=src_stream,
                                         layout=arg['layout'],
                                         reader_factory=rf)

    return pipes_in, pipes_out


async def build_network_pipes(inputs: Dict[str, str],
                              outputs: Dict[str, str],
                              configured_streams: Dict[str, stream.Stream],
                              my_node: BaseNode,
                              start_time: Optional[int],
                              end_time: Optional[int],
                              force=False):
    if not force:
        _display_warning(outputs.values(), start_time, end_time)

    pipes_in = {}
    pipes_out = {}
    try:
        for name in inputs:
            my_stream = await _parse_stream(my_node, inputs[name], configured_streams)
            if start_time is None and end_time is None:
                # subscribe to live data
                pipes_in[name] = await my_node.data_subscribe(my_stream)
            else:
                pipes_in[name] = await my_node.data_read(my_stream,
                                                         start_time,
                                                         end_time)

        for name in outputs:
            my_stream = await _parse_stream(my_node, outputs[name], configured_streams)
            pipes_out[name] = await my_node.data_write(my_stream,
                                                       start_time,
                                                       end_time)
    except (errors.ApiError, errors.ConfigurationError) as e:
        # close any pipes that were created
        for name in pipes_in:
            await pipes_in[name].close()
        for name in pipes_out:
            await pipes_out[name].close()
        # re-raise the exception to be handled elsewhere
        raise e

    return pipes_in, pipes_out


def _display_warning(paths, start_time, end_time):
    # warn about data removal for historic execution
    if start_time is not None or end_time is not None:
        if end_time is None:
            msg = "after [%s]" % timestamp_to_human(start_time)
        elif start_time is None:
            msg = "before [%s]" % timestamp_to_human(end_time)
        else:
            msg = "between [%s - %s]" % (timestamp_to_human(start_time),
                                         timestamp_to_human(end_time))
        output_paths = ", ".join([x.split(':')[0] for x in paths])
        if not click.confirm("This will remove any data %s in the output streams [%s]" % (
                msg, output_paths)):
            exit(1)


async def _parse_stream(node: BaseNode, pipe_config, configured_streams: Dict[str, stream.Stream]) -> stream.Stream:
    (path, name, inline_config) = parse_pipe_config(pipe_config)
    full_path = "/".join([path, name])
    configured_stream = None
    # inline config exists
    if inline_config != "":
        (datatype, element_names) = parse_inline_config(inline_config)
        datatype = datatype.name.lower()  # API models are plain text attributes
        # make sure inline config agrees with stream config if present
        if path in configured_streams.keys():
            configured_stream = configured_streams[path]
            if datatype != configured_stream.datatype.lower() or \
                    len(configured_stream.elements) != len(element_names):
                raise errors.ConfigurationError("Invalid configuration: [%s] inline format does not match config file" %
                                                name)
        # otherwise create a stream object from the inline config
        configured_stream = stream.Stream()
        configured_stream.name = name
        configured_stream.decimate = True
        configured_stream.datatype = datatype
        for i in range(len(element_names)):
            e = stream.Element()
            e.name = element_names[i]
            e.index = i
            configured_stream.elements.append(e)
    # no inline config
    elif full_path in configured_streams.keys():
        configured_stream = configured_streams[full_path]

    # use API to get or create the stream on the Joule node
    try:
        remote_stream = await node.stream_get(path + '/' + name)
        if configured_stream is not None:
            # make sure the provided config matches the existing stream
            if remote_stream.datatype != configured_stream.datatype or \
                    len(remote_stream.elements) != len(configured_stream.elements):
                raise errors.ConfigurationError("Invalid stream configuration: [%s] has layout: %s not %s_%d" %
                                                (remote_stream.name, remote_stream.layout,
                                                 configured_stream.datatype,
                                                 len(configured_stream.elements)))
    except errors.ApiError as e:
        if '404' in str(e):
            if configured_stream is None:
                raise errors.ConfigurationError(
                    "[%s] is invalid: must configure inline or stream configuration file" % pipe_config)

            log.info("creating output stream [%s%s]" % (path, name))
            remote_stream = await node.stream_create(configured_stream,
                                                     path)
        else:
            raise e
    return remote_stream



import numpy as np
import asyncio
import logging

from joule.models.pipes import Pipe
from joule.models.pipes.errors import PipeError, EmptyPipe

Loop = asyncio.AbstractEventLoop
log = logging.getLogger('joule')


class LocalPipe(Pipe):
    """
           Pipe for intra-module communication.

           Args:
               layout: ``datatype_width``, for example ``float32_3`` for a three element stream
                 must. See Stream.layout
           Keyword Args:
               loop: specify a an event loop, otherwise the default one is used
               name: useful for debugging with multiple pipes
               close_cb: callback coroutine executed when pipe closes
               debug: enable to log pipe usage events
    """

    def __init__(self, layout: str, loop: Loop = None, name: str = None,
                 close_cb = None, debug: bool = False, stream=None, write_limit=0):

        super().__init__(name=name, layout=layout, stream=stream)
        if loop is None:
            loop = asyncio.get_event_loop()
        # tunable constants
        self.TIMEOUT_INTERVAL = 0.5
        self.debug = debug
        self.interval_break = False
        self.closed = False
        self.read_buffer = np.empty((0,), dtype=self.dtype)
        self.close_cb = close_cb
        self._reread = False
        # initialize buffer and queue
        self.queue = asyncio.Queue(loop=loop, maxsize=write_limit)
        self.queued_rows = 0
        self.last_index = 0
        self.direction = Pipe.DIRECTION.TWOWAY
        # caching
        self._caching = False
        self._cache_index = 0
        self._cache = None

    async def read(self, flatten=False):
        if self._failed:
            await self.close()
            raise PipeError('pipe failed')

        # if reread is set just return the old data
        if self._reread:
            self._reread = False
            if len(self.read_buffer) == 0:
                raise PipeError("No data left to reread")
            return self._format_data(self.read_buffer, flatten)

        self.interval_break = False

        # if the queue is empty and we have old data, just return the old data
        if self.queue.empty() and len(self.read_buffer) > 0:
            return self._format_data(self.read_buffer, flatten)

        # otherwise wait for at least one block
        while self.queue.empty():
            # if the buffer is empty and the queue is empty and the pipe is closed
            if self.queue.empty() and len(self.read_buffer) == 0 and self.closed:
                raise EmptyPipe()
            await asyncio.sleep(self.TIMEOUT_INTERVAL)

        return self._read(flatten)

    def read_nowait(self, flatten=False):
        """
        Same as read but this is not a coroutine. This should only be used for unit testing.

        Args:
            flatten:

        Returns:
            numpy.ndarray

        >>> data = pipe.read_nowait()
        [1, 2, 3]

        """
        if self._failed:
            raise PipeError('pipe failed')

        # if reread is set just return the old data
        if self._reread:
            self._reread = False
            if len(self.read_buffer) == 0:
                raise PipeError("No data left to reread")
            return self._format_data(self.read_buffer, flatten)

        # if the queue is empty and we have old data, just return the old data
        if self.queue.empty() and len(self.read_buffer) > 0:
            return self._format_data(self.read_buffer, flatten)

        # if the buffer is empty and the queue is empty and the pipe is closed
        if self.queue.empty() and len(self.read_buffer) == 0 and self.closed:
            raise EmptyPipe()
        # do not wait for new data, return an empty array if nothing else is available
        return self._read(flatten)

    def _read(self, flatten=False):
        # now put all the queued data together in a single array with the previous data
        # this cannot be interrupted, relies on the total size of data written to the pipe
        start = 0
        end = len(self.read_buffer)
        buffer = np.empty((self.queued_rows + end,), self.dtype)
        if self.debug:
            print("[%s:read] initialized %d row buffer" % (self.name, len(buffer)))
            print("[%s:read] adding %d rows of unconsumed data" % (self.name, len(self.read_buffer)))
        buffer[start:end] = self.read_buffer
        start = end

        while not self.queue.empty():
            block = self.queue.get_nowait()
            if block is None:
                self.interval_break = True
                break
            end = start + len(block)
            buffer[start:end] = block
            start = end
            self.queued_rows -= len(block)

        self.read_buffer = buffer[:end]
        if self.debug:
            print("[%s:read] returning %d rows" % (self.name, len(self.read_buffer)))
        return self._format_data(self.read_buffer, flatten)

    def reread_last(self):
        if len(self.read_buffer) == 0:
            raise PipeError("No data left to reread")
        self._reread = True

    @property
    def end_of_interval(self):
        return self.interval_break

    def consume(self, num_rows):
        if num_rows == 0:
            return
        if num_rows < 0:
            raise PipeError("consume called with negative offset: %d" % num_rows)
        if num_rows > len(self.read_buffer):
            raise PipeError("cannot consume %d rows: only %d available"
                            % (num_rows, len(self.read_buffer)))
        if self.debug:
            print("[%s:read] consumed %d rows" % (self.name, num_rows))
        self.read_buffer = self.read_buffer[num_rows:]

    async def write(self, data: np.ndarray):
        if self._failed:
            await self.close()
            raise PipeError('pipe failed')
        if self.closed:
            raise PipeError("Cannot write to a closed pipe")
        if not self._validate_data(data):
            return
        # convert into a structured array
        sarray = self._apply_dtype(data)

        if self._caching:
            for row in sarray:
                self._cache[self._cache_index] = row
                self._cache_index += 1
                if self._cache_index >= len(self._cache):
                    await self.flush_cache()
        else:
            await self._write(sarray)

    async def _write(self, sarray):
        # send data to subscribers
        for pipe in self.subscribers:
            await pipe.write(sarray)

        # if the queue size is infinite do not wait
        if self.queue.maxsize <= 0:
            self.queue.put_nowait(sarray)
        else:
            # wait until a slot is available
            await self.queue.put(sarray)
        self.queued_rows += len(sarray)
        await asyncio.sleep(0)
        if self.debug:
            print("[%s:write] queueing block with [%d] rows" % (self.name, len(sarray)))

    def write_nowait(self, data):
        if self._failed:
            raise PipeError('pipe failed')
        if self.closed:
            raise PipeError("Cannot write to a closed pipe")
        if not self._validate_data(data):
            return

        # convert into a structured array
        sarray = self._apply_dtype(data)
        # send data to subscribers
        for pipe in self.subscribers:
            if type(pipe) is LocalPipe:
                p: LocalPipe = pipe  # to appease type checker
                p.write_nowait(sarray)
            else:
                raise PipeError("cannot write_nowait to subscriber [%s]" % pipe.name)

        self.queue.put_nowait(sarray)
        self.queued_rows += len(sarray)
        if self.debug:
            print("[%s:write] queueing block with [%d] rows" % (self.name, len(sarray)))

    def enable_cache(self, lines: int):
        self._caching = True
        self._cache = np.empty(lines, self.dtype)
        self._cache_index = 0

    async def flush_cache(self):
        if self.closed:
            raise PipeError("Cannot write to a closed pipe")
        if self._cache_index > 0:
            await self._write(self._cache[:self._cache_index])
            self._cache_index = 0
            self._cache = np.empty(len(self._cache), self.dtype)

    async def close_interval(self):
        if self._failed:
            raise PipeError('pipe failed')
        if self.closed:
            raise PipeError("Cannot write to a closed pipe")
        if self.debug:
            print("[%s:write] closing interval" % self.name)
        if self._caching:
            await self.flush_cache()
        await self.queue.put(None)

    def close_interval_nowait(self):
        """
        Same as close_interval but this is not a coroutine. This should only be used for
        unit testing

        """
        if self._failed:
            raise PipeError('pipe failed')
        if self.debug:
            print("[%s:write] closing interval" % self.name)
        self.queue.put_nowait(None)

    def change_layout(self, layout: str):
        self._layout = layout
        self.read_buffer = np.empty((0,), dtype=self.dtype)
        self.queued_rows = 0
        self.last_index = 0
        # caching
        if self._caching:
            self._cache = np.empty(len(self._cache), self.dtype)
            self._cache_index = 0

    async def close(self):
        if self.closed:
            return
        if self._caching:
            await self.flush_cache()
        # close any subscribers
        for pipe in self.subscribers:
            await pipe.close()
        self.closed = True
        if self.close_cb is not None:
            await self.close_cb()

    def close_nowait(self):
        """
        Same as close but this is not a coroutine. This should only be used for
        unit testing

        """
        if len(self.subscribers) > 0:
            raise PipeError("cannot close_nowait subscribers, use async")
        if self.close_cb is not None:
            raise PipeError("close_cb cannot be executed, use async")
        self.closed = True




from aiohttp.web import middleware, HTTPForbidden, HTTPBadRequest
from sqlalchemy.orm import Session
import logging
from sqlalchemy.exc import SQLAlchemyError
from joule.models import Master

log = logging.getLogger('joule')


def authorize(exemptions=None):
    if exemptions is None:
        exemptions = []

    @middleware
    async def _authorize(request, handler):
        db: Session = request.app["db"]

        # Remote is empty for Unix Socket connections,
        # populate app with values from reverse proxy if present
        if request.remote == "":
            # set the app port and scheme based off headers if present
            if 'X-API-PORT' in request.headers:
                request.app["port"] = request.headers['X-API-PORT']
            if 'X-API-SCHEME' in request.headers:
                request.app["scheme"] = request.headers['X-API-SCHEME']
            if 'X-API-BASE-URI' in request.headers:
                request.app["base_uri"] = request.headers['X-API-BASE-URI']
            if 'X-FORWARDED-FOR' in request.headers:
                request.app["remote_ip"] = request.headers['X-FORWARDED-FOR']
            # OK, skip authorization unless requested (ie this is from a reverse proxy)
            if (('X-AUTH-REQUIRED' not in request.headers) or
               ([request.method, request.path] in exemptions)):
                return await handler(request)
        # This is not coming through a proxy, populate with locally "true" values
        else:
            request.app["base_uri"] = ""
            request.app["remote_ip"] = request.remote
            # scheme and port are populated already (by the daemon)

        # OK, exempt request
        if [request.method, request.path] in exemptions:
            return await handler(request)
        # ERROR: Missing Key
        if 'X-API-KEY' not in request.headers:
            raise HTTPForbidden()
        key = request.headers['X-API-KEY']
        # ERROR: Invalid Key
        if (db.query(Master).
                filter(Master.key == key).
                count() != 1):
            raise HTTPForbidden
        # OK, valid key
        return await handler(request)

    return _authorize


@middleware
async def sql_rollback(request, handler):
    db: Session = request.app["db"]
    try:
        return await handler(request)
    except SQLAlchemyError as e:
        log.warning("Invalid HTTP request: %s", e)
        db.rollback()
        raise HTTPBadRequest



import click
from .copy import data_copy
from .read import cmd as data_read
from .remove import data_remove
from .intervals import intervals
from .consolidate import consolidate


@click.group(name="data")
def data():
    """Manage Joule data"""
    pass  # pragma: no cover


data.add_command(data_copy)
data.add_command(data_read)
data.add_command(data_remove)
data.add_command(intervals)
data.add_command(consolidate)


import asyncio
from typing import List, Optional, Callable, Coroutine, Dict
import numpy as np
from io import BytesIO
import asyncpg
import asyncpg.exceptions
import datetime
import shutil
from joule.utilities import interval_tools, timestamp_to_datetime
from joule.models.data_store.timescale_inserter import Inserter
from joule.models.data_store import psql_helpers
from joule.models import Stream, pipes
from joule.models.data_store.data_store import DataStore, StreamInfo, DbInfo

Loop = asyncio.AbstractEventLoop


class TimescaleStore(DataStore):

    def __init__(self, dsn: str, insert_period: float,
                 cleanup_period: float, loop: Loop):
        self.dsn = dsn
        self.decimation_factor = 4
        self.insert_period = insert_period
        self.cleanup_period = cleanup_period
        self.loop = loop
        self.pool = None
        # tunable constants
        self.extract_block_size = 50000

    async def initialize(self, streams: List[Stream]) -> None:
        self.pool = await asyncpg.create_pool(self.dsn, command_timeout=60)

    async def close(self):
        await self.pool.close()

    async def spawn_inserter(self, stream: 'Stream', pipe: pipes.Pipe,
                             loop: Loop, insert_period=None) -> asyncio.Task:
        if insert_period is None:
            insert_period = self.insert_period

        inserter = Inserter(self.pool, stream,
                            insert_period, self.cleanup_period)
        return loop.create_task(inserter.run(pipe))

    async def insert(self, stream: 'Stream',
                     data: np.ndarray, start: int, end: int):
        pass

    async def consolidate(self, stream: 'Stream', start: int, end: int, max_gap: int) -> int:
        # remove interval gaps less than or equal to max_gap duration (in us)
        intervals = await self.intervals(stream, start, end)
        if len(intervals) == 0:
            return  # no data, nothing to do
        duration = [intervals[0][0], intervals[-1][1]]

        gaps = interval_tools.interval_difference([duration], intervals)

        if len(gaps) == 0:
            return  # no interval breaks, nothing to do
        small_gaps = [gap for gap in gaps if (gap[1] - gap[0]) <= max_gap]
        boundaries = [gap[0] for gap in small_gaps]
        str_datetimes = ["'%s'" % str(timestamp_to_datetime(ts)) for ts in boundaries]
        query = "DELETE FROM data.stream%d_intervals WHERE time IN (%s)" % (stream.id, ",".join(str_datetimes))
        async with self.pool.acquire() as conn:
            await conn.execute(query)
        return len(small_gaps)

    async def intervals(self, stream: 'Stream', start: Optional[int], end: Optional[int]):
        # find the intervals
        # for each interval find the nearest data
        intervals = []
        async with self.pool.acquire() as conn:
            boundaries = await psql_helpers.get_boundaries(conn, stream, start, end)
            cur_start = None
            for i in range(len(boundaries)):
                boundary = boundaries[i]
                if cur_start is not None:
                    query = "SELECT time FROM data.stream%d WHERE time < '%s'" % (stream.id, boundary)
                    query += " ORDER BY time DESC LIMIT 1"
                    prev_ts = await conn.fetchval(query)
                    utc_start_ts = int(cur_start.replace(tzinfo=datetime.timezone.utc).timestamp() * 1e6)
                    utc_end_ts = int(prev_ts.replace(tzinfo=datetime.timezone.utc).timestamp() * 1e6)
                    # intervals are [..) with extra us on the end
                    intervals.append([utc_start_ts, utc_end_ts + 1])
                    cur_start = None
                query = "SELECT time FROM data.stream%d WHERE time >= '%s'" % (stream.id, boundary)
                if i < (len(boundaries) - 1):
                    query += " AND time < '%s' " % (boundaries[i + 1])
                query += " ORDER BY time ASC LIMIT 1"
                next_ts = await conn.fetchval(query)
                if next_ts is None:
                    # unecessary interval marker, remove it?
                    continue
                if cur_start is None:
                    cur_start = next_ts
        return intervals

    async def extract(self, stream: 'Stream', start: Optional[int], end: Optional[int],
                      callback: Callable[[np.ndarray, str, int], Coroutine],
                      max_rows: int = None, decimation_level=None):
        conn = await self.pool.acquire()
        # limit time bounds to range of base stream
        (start, end) = await psql_helpers.limit_time_bounds(conn, stream, start, end)
        # figure out appropriate decimation level
        if decimation_level is None:
            if max_rows is None:
                decimation_level = 1
            else:
                # find out how much data this represents
                count = await psql_helpers.get_row_count(conn, stream, start, end)
                if count > 0:
                    desired_decimation = np.ceil(count / max_rows)

                    decimation_level = int(4 ** np.ceil(np.log(desired_decimation) /
                                                        np.log(self.decimation_factor)))
                    # print("count=%d, max_rows=%d,desired_decim=%d,decim_level=%d" % (
                    #    count, max_rows, desired_decimation, decimation_level))
                else:
                    # create an empty array with the right data type
                    data = np.array([], dtype=pipes.compute_dtype(stream.layout))
                    await callback(data, stream.layout, 1)
                    await self.pool.release(conn)
                    return
        try:
            await _extract_data(conn, stream, callback, decimation_level, start, end,
                                block_size=self.extract_block_size)
        except Exception as e:
            raise e
        finally:
            await self.pool.release(conn)

    async def remove(self, stream: 'Stream',
                     start: Optional[int], end: Optional[int],
                     exact: bool = True):
        where_clause = psql_helpers.query_time_bounds(start, end)
        async with self.pool.acquire() as conn:
            tables = await psql_helpers.get_table_names(conn, stream)
            for table in tables:
                # TODO: use drop chunks with newer and older clauses when timescale is updated
                # ******DROP CHUNKS IS *VERY* APPROXIMATE*********
                if start is None and "intervals" not in table and not exact:
                    # use the much faster drop chunks utility and accept the approximate result
                    bounds = await psql_helpers.convert_time_bounds(conn, stream, start, end)
                    if bounds is None:
                        return  # no data to remove
                    query = "SELECT drop_chunks(table_name=>'%s', schema_name=>'data', older_than=>'%s'::timestamp)" % \
                            (table.split(".")[1], bounds[1])
                else:
                    query = 'DELETE FROM %s ' % table + where_clause
                try:
                    await conn.execute(query)
                except asyncpg.UndefinedTableError:
                    return  # no data to remove
                except asyncpg.exceptions.RaiseError as err:
                    print("psql: ", err)
                    return
            # create an interval boundary to mark the missing data
            if start is not None:
                await psql_helpers.close_interval(conn, stream, start)

    async def destroy(self, stream: 'Stream'):
        async with self.pool.acquire() as conn:
            tables = await psql_helpers.get_table_names(conn, stream)
            for table in tables:
                try:
                    await conn.execute('DROP TABLE %s ' % table)
                except asyncpg.UndefinedTableError:
                    pass

    async def destroy_all(self):
        async with self.pool.acquire() as conn:
            tables = await psql_helpers.get_all_table_names(conn)
            for table in tables:
                try:
                    await conn.execute('DROP TABLE %s ' % table)
                except asyncpg.UndefinedTableError:
                    pass

    async def info(self, streams: List['Stream']) -> Dict[int, StreamInfo]:
        results = {}
        async with self.pool.acquire() as conn:
            for my_stream in streams:
                try:
                    record = await conn.fetchrow("select * from stream_info(%d);" % my_stream.id)
                except asyncpg.UndefinedTableError:
                    # this stream has no data tables, just make up an empty record
                    record = {'min_ts': None, 'max_ts': None, 'rows': 0, 'size': 0}
                start = record['min_ts']
                end = record['max_ts']

                if start is not None:
                    start = int(start.replace(tzinfo=datetime.timezone.utc)
                                .timestamp() * 1e6)
                if end is not None:
                    end = int(end.replace(tzinfo=datetime.timezone.utc)
                              .timestamp() * 1e6)
                if start is not None and end is not None:
                    total_time = end - start
                else:
                    total_time = 0
                results[my_stream.id] = StreamInfo(start, end,
                                                   record['rows'],
                                                   total_time, record['size'])
        return results

    async def dbinfo(self) -> DbInfo:

        async with self.pool.acquire() as conn:
            dbsize = await conn.fetchval("select pg_database_size(current_database())")  # in bytes
            path = await conn.fetchval("show data_directory")
            try:
                usage = shutil.disk_usage(path)  # usage in bytes
                return DbInfo(path=path,
                              size=dbsize,
                              other=max(usage.used - dbsize, 0),
                              reserved=max(usage.total - usage.used - usage.free, 0),
                              free=usage.free)
            except FileNotFoundError:
                # this occurs if the database is remote
                return DbInfo(path="--remote-database--",
                              size=-1,
                              other=-1,
                              reserved=-1,
                              free=-1)


async def _extract_data(conn: asyncpg.Connection, stream: Stream, callback,
                        decimation_level: int = 1, start: int = None, end: int = None,
                        block_size=50000):
    if decimation_level > 1:
        layout = stream.decimated_layout
    else:
        layout = stream.layout

    table_name = "data.stream%d" % stream.id
    if decimation_level > 1:
        table_name += "_%d" % decimation_level
    # extract by interval
    query = "SELECT time FROM data.stream%d_intervals " % stream.id
    query += psql_helpers.query_time_bounds(start, end)
    try:
        boundary_records = await conn.fetch(query)
    except asyncpg.UndefinedTableError:
        # no data tables
        data = np.array([], dtype=pipes.compute_dtype(layout))
        await callback(data, layout, decimation_level)
        return

    boundary_records += [{'time': end}]
    for i in range(len(boundary_records)):
        record = boundary_records[i]
        end = record['time']
        # extract the interval data
        done = False
        while not done:
            query = "SELECT * FROM %s " % table_name
            query += psql_helpers.query_time_bounds(start, end)
            query += " ORDER BY time ASC LIMIT %d" % block_size
            psql_bytes = BytesIO()
            try:
                await conn.copy_from_query(query, format='binary', output=psql_bytes)
            except asyncpg.UndefinedTableError:
                # interval table exists but not the data table
                data = np.array([], dtype=pipes.compute_dtype(layout))
                await callback(data, layout, decimation_level)
                return
            psql_bytes.seek(0)
            dtype = pipes.compute_dtype(layout)
            np_data = psql_helpers.bytes_to_data(psql_bytes, dtype)
            await callback(np_data, layout, decimation_level)

            if len(np_data) < block_size:
                break
            start = np_data['timestamp'][-1] + 1
        # do not put an interval token at the end of the data
        if i < len(boundary_records) - 1:
            await callback(pipes.interval_token(layout), layout, decimation_level)
        start = end



import asyncio
import numpy as np
import random
import logging
import asyncpg

from joule.models import Stream, pipes
import joule.utilities
from joule.models.data_store import psql_helpers

Loop = asyncio.AbstractEventLoop
log = logging.getLogger('joule')


class Inserter:

    def __init__(self, pool: asyncpg.pool.Pool, stream: Stream, insert_period: float,
                 cleanup_period: float):
        self.pool = pool
        self.stream = stream
        # round cleanup_period to a multiple of insert_period
        if insert_period == 0 or cleanup_period < insert_period:
            self.cleanup_interval = 1  # clean with every insert
        else:
            self.cleanup_interval = np.ceil(cleanup_period / insert_period)
        # add offsets to the period to distribute traffic
        self.insert_period = insert_period + insert_period * random.random() * 0.5
        self.decimator: Decimator = None

    async def run(self, pipe: pipes.Pipe) -> None:
        """insert stream data from the queue until the queue is empty"""

        # lazy stream creation
        try:
            async with self.pool.acquire() as conn:
                await psql_helpers.create_stream_table(conn, self.stream)
        except asyncio.CancelledError:
            return

        # close the beginning of the data insert
        first_insert = True
        # track the last timestamp inserted
        last_ts = None
        ticks = 0
        try:
            while True:
                await asyncio.sleep(self.insert_period)
                data = await pipe.read()
                # there might be an interval break and no new data
                async with self.pool.acquire() as conn:
                    if len(data) > 0:
                        if first_insert:
                            first_insert = False
                            await psql_helpers.close_interval(conn, self.stream, data['timestamp'][0] - 1)
                        last_ts = data['timestamp'][-1]
                        # lazy initialization of decimator
                        if self.stream.decimate and self.decimator is None:
                            self.decimator = Decimator(self.stream, 1, 4)
                        psql_bytes = psql_helpers.data_to_bytes(data)
                        await conn.copy_to_table("stream%d" % self.stream.id,
                                                 schema_name='data',
                                                 format='binary',
                                                 source=psql_bytes)
                        # this was successful so consume the data
                        pipe.consume(len(data))
                        # decimate the data
                        if self.decimator is not None:
                            await self.decimator.process(conn, data)
                    # check for interval breaks
                    if pipe.end_of_interval and last_ts is not None:
                        await psql_helpers.close_interval(conn, self.stream, last_ts)
                        if self.decimator is not None:
                            self.decimator.close_interval()
                    ticks += 1
                    if ticks % self.cleanup_interval == 0:
                        await self.cleanup(conn)

        except (pipes.EmptyPipe, asyncio.CancelledError):
            pass

    async def cleanup(self, conn: asyncpg.Connection):
        if self.stream.keep_us == Stream.KEEP_ALL:
            return
        tables = await psql_helpers.get_table_names(conn, self.stream, with_schema=False)
        # ts is milliseconds UNIX timestamp
        keep_s = self.stream.keep_us // 1e6
        for table in tables:
            if 'interval' in table:
                # drop all boundaries before the cutoff
                cutoff = joule.utilities.timestamp_to_datetime(
                    joule.utilities.time_now() - self.stream.keep_us)
                query = "DELETE FROM data.%s WHERE time < '%s'" % (table, cutoff)
            else:
                query = "SELECT drop_chunks(interval '%d seconds', '%s', 'data')" % (keep_s, table)
            await conn.execute(query)


class Decimator:

    def __init__(self, stream: Stream, from_level: int, factor: int, debug=False):
        self.stream = stream
        self.level = from_level * factor
        self.table_name = "stream%d_%d" % (stream.id, self.level)
        self.full_table_name = "data." + self.table_name
        if from_level > 1:
            self.again = True
        else:
            self.again = False
        self.factor = factor
        self.layout = stream.decimated_layout
        self.buffer = []
        self.path_created = False
        self.child: Decimator = None
        self.debug = debug
        if self.debug:
            print("creating decim level %d" % self.level)
        # hold off to rate limit traffic
        self.holdoff = 0  # random.random()

    async def process(self, conn: asyncpg.Connection, data: np.ndarray) -> None:
        """decimate data and insert it, retry on error"""
        if not self.path_created:
            await psql_helpers.create_decimation_table(conn, self.stream, self.level)
            self.path_created = True

        decim_data = self._process(data)
        if self.debug:
            print("\t level %d: %d rows" % (self.level, len(decim_data)))
        if len(decim_data) == 0:
            return

        # lazy initialization of child
        if self.child is None:
            self.child = Decimator(self.stream, self.level, self.factor)

        psql_bytes = psql_helpers.data_to_bytes(decim_data)
        await conn.copy_to_table(self.table_name,
                                 schema_name='data',
                                 format='binary',
                                 source=psql_bytes)
        await self.child.process(conn, decim_data)

    def close_interval(self):
        self.buffer = []
        if self.child is not None:
            self.child.close_interval()

    def _process(self, sarray: np.ndarray) -> np.ndarray:

        # flatten structured array
        data = np.c_[sarray['timestamp'][:, None], sarray['data']]

        # check if there is old data
        if len(self.buffer) != 0:
            # append the new data onto the old data
            data = np.concatenate((self.buffer, data))
        (n, m) = data.shape

        # Figure out which columns to use as the input for mean, min, and max,
        # depending on whether this is the first decimation or we're decimating
        # again.  Note that we include the timestamp in the means.
        if self.again:
            c = (m - 1) // 3
            # e.g. c = 3
            # ts mean1 mean2 mean3 min1 min2 min3 max1 max2 max3
            mean_col = slice(0, c + 1)
            min_col = slice(c + 1, 2 * c + 1)
            max_col = slice(2 * c + 1, 3 * c + 1)
        else:
            mean_col = slice(0, m)
            min_col = slice(1, m)
            max_col = slice(1, m)

        # Discard extra rows that aren't a multiple of factor
        n = n // self.factor * self.factor

        if n == 0:  # not enough data to work with, save it for later
            self.buffer = data
            return np.array([])

        trunc_data = data[:n, :]
        # keep the leftover data
        self.buffer = np.copy(data[n:, :])

        # Reshape it into 3D so we can process 'factor' rows at a time
        trunc_data = trunc_data.reshape(n // self.factor, self.factor, m)

        # Fill the result
        out = np.c_[np.mean(trunc_data[:, :, mean_col], axis=1),
                    np.min(trunc_data[:, :, min_col], axis=1),
                    np.max(trunc_data[:, :, max_col], axis=1)]

        # structure the array
        width = np.shape(out)[1] - 1
        dtype = np.dtype([('timestamp', '<i8'), ('data', '<f4', width)])
        sout = np.zeros(out.shape[0], dtype=dtype)
        sout['timestamp'] = out[:, 0]
        sout['data'] = out[:, 1:]
        # insert the data into the database
        return sout



import argparse
import asyncio
import signal

from typing import List, Tuple, Dict, Optional
from aiohttp import web
import socket
import logging
import os
import uvloop

from joule.api import node
from joule import api
from joule.client import helpers
# import directly so it can be mocked easily in unit tests
from joule.errors import ConfigurationError, EmptyPipeError
from joule.models import pipes, stream

Pipes = Dict[str, 'pipes.Pipe']
Loop = asyncio.AbstractEventLoop
log = logging.getLogger('joule')


class BaseModule:
    """
    This is an abstract class and should not be inherited directly. Instead inherit from one of
    the following children :class:`joule.ReaderModule`, :class:`joule.FilterModule`, or :class:`joule.CompositeModule`
    """

    def __init__(self):
        self.stop_requested = False
        self.pipes: List[pipes.Pipe] = []
        self.STOP_TIMEOUT = 2
        self.runner = None
        self.node = None

    async def run_as_task(self, parsed_args, app: web.Application, loop) -> asyncio.Task:
        assert False, "implement in child class"  # pragma: no cover

    def custom_args(self, parser: argparse.ArgumentParser):
        """

        Override to add custom command line arguments to the module.

        .. code-block:: python

            class ModuleDemo(BaseModule):

                def custom_args(self, parser):
                     parser.description = "**module description**"
                     # add optional help text to the argument
                     parser.add_argument("--arg", help="custom argument")
                     # parse json input
                     parser.add_argument("--json_arg", type=json.loads)
                     # a yes|no argument that resolves to True|False
                     parser.add_argument("--flag_arg", type=joule.yesno)

                #... other module code


        Always use keyword arguments with modules so they can be specified
        in the **[Arguments]** section  of module configuration file

        Use the ``type`` parameter to specify a parser function. The parser
        function should accept a string input and return the appropriate object.

        """
        # parser.add_argument("--custom_flag")
        pass  # pragma: no cover

    def stop(self):
        """
        Override to change the default shutdown strategy which simply sets
        the ``stop_requested`` flag. If a module does not terminate within a few seconds
        of this method being called Joule will forcibly stop the module with SIGKILL.
        """
        # override in client for alternate shutdown strategy
        self.stop_requested = True

    def create_dev_app(self, loop: asyncio.AbstractEventLoop) -> web.Application:
        parser = argparse.ArgumentParser()
        self._build_args(parser)
        # must specify module config in JOULE_MODULE_CONFIG environment variable
        if 'JOULE_MODULE_CONFIG' not in os.environ:
            raise ConfigurationError("JOULE_MODULE_CONFIG not set, must specify config file")
        module_config_file = os.environ['JOULE_MODULE_CONFIG']
        if not os.path.isfile(module_config_file):
            raise ConfigurationError("JOULE_MODULE_CONFIG is not a module config file")

        module_args = helpers.load_args_from_file(module_config_file)
        parsed_args = parser.parse_args(module_args)
        self.node = api.get_node(parsed_args.node)

        self.stop_requested = False
        my_app = self._create_app()

        async def on_startup(app):
            app['task'] = await self.run_as_task(parsed_args, app, loop)

        my_app.on_startup.append(on_startup)

        async def on_shutdown(app):
            self.stop()
            await app['task']

        my_app.on_shutdown.append(on_shutdown)
        return my_app

    def start(self, parsed_args: argparse.Namespace = None):
        """
        Execute the module. Do not override this function. Creates an event loop and
        executes the :meth:`run` coroutine.

        Args:
            parsed_args: omit to parse the command line arguments

        .. code-block:: python

            class ModuleDemo(BaseModule):
                # body of module...
                # at a minimum the run coroutine must be implemented

            if __name__ == "__main__":
                my_module = ModuleDemo()
                my_module.start()
        """

        if parsed_args is None:  # pragma: no cover
            parser = argparse.ArgumentParser()
            self._build_args(parser)
            module_args = helpers.module_args()
            parsed_args = parser.parse_args(module_args)

        #asyncio.set_event_loop_policy(uvloop.EventLoopPolicy())
        loop = asyncio.get_event_loop()
        self.stop_requested = False
        if parsed_args.api_socket != "unset":
            # should be set by the joule daemon
            if 'JOULE_CA_FILE' in os.environ:
                cafile = os.environ['JOULE_CA_FILE']
            else:
                cafile = ""
            self.node = node.UnixNode("local", parsed_args.api_socket, cafile, loop)
        else:
            self.node = api.get_node(parsed_args.node)
        self.runner: web.AppRunner = loop.run_until_complete(self._start_interface(parsed_args))
        app = None
        if self.runner is not None:
            app = self.runner.app
        try:
            task = loop.run_until_complete(
                self.run_as_task(parsed_args, app, loop))
        except ConfigurationError as e:
            log.error("ERROR: " + str(e))
            self._cleanup(loop)
            return

        def stop_task():
            # give task no more than 2 seconds to exit
            loop.call_later(self.STOP_TIMEOUT, task.cancel)
            # run custom exit routine
            self.stop()

        loop.add_signal_handler(signal.SIGINT, stop_task)
        loop.add_signal_handler(signal.SIGTERM, stop_task)
        try:
            loop.run_until_complete(task)
        except (asyncio.CancelledError, pipes.EmptyPipe, EmptyPipeError):
            pass
        self._cleanup(loop)

    def _cleanup(self, loop: Loop):

        if self.runner is not None:
            loop.run_until_complete(self.runner.cleanup())
        for pipe in self.pipes:
            loop.run_until_complete(pipe.close())
        if self.node is not None:
            loop.run_until_complete(self.node.close())
        loop.run_until_complete(asyncio.sleep(0))
        loop.close()

    def _build_args(self, parser):  # pragma: no cover
        grp = parser.add_argument_group('joule',
                                        'control module execution')
        # --pipes: JSON argument set by jouled
        grp.add_argument("--pipes",
                         default="unset",
                         help='RESERVED, managed by jouled')
        # --socket: UNIX socket set by jouled for interface proxy
        grp.add_argument("--socket",
                         default="unset",
                         help='RESERVED, managed by jouled')
        # --api-socket: UNIX socket for API calls back to jouled
        grp.add_argument("--api-socket",
                         default="unset",
                         help='RESERVED, managed by jouled')
        # --port: port to host interface on during isolated execution
        grp.add_argument("--port", type=int,
                         default=8080,
                         help='port for isolated execution')
        # --host: IP address to host interface on during isolated exedcution
        grp.add_argument("--host",
                         default="0.0.0.0",
                         help="IP address for isolated execution")
        # --module_config: set to run module standalone
        grp.add_argument("--module_config",
                         default="unset",
                         help='specify *.conf file for isolated execution')
        # --stream_configs: set to run module standalone
        grp.add_argument("--stream_configs",
                         default="unset",
                         help="specify directory of stream configs " +
                              "for isolated execution")
        # --node: node to connect to for isolated execution
        grp.add_argument("--node", default="",
                         help="joule node for isolated execution")
        # --start_time: historical isolation mode
        grp.add_argument("--start_time", default=None,
                         help="input start time for historic isolation")
        # --end_time: historical isolation mode
        grp.add_argument("--end_time", default=None,
                         help="input end time for historic isolation")
        # --force: do not ask for confirmation
        grp.add_argument("--force", action="store_true",
                         help="do not ask for confirmation before dangerous operations")

        parser.formatter_class = argparse.RawDescriptionHelpFormatter
        self.custom_args(parser)

    async def _build_pipes(self, parsed_args) -> Tuple[Pipes, Pipes]:
        # for unit testing mocks
        from joule.client.helpers.pipes import build_network_pipes

        pipe_args = parsed_args.pipes

        # figure out whether we should run with fd's or network sockets
        if pipe_args == 'unset':
            start, end = helpers.validate_time_bounds(parsed_args.start_time,
                                                      parsed_args.end_time)
            if parsed_args.module_config == 'unset':
                raise ConfigurationError('must specify --module_config')
            inputs = {}
            outputs = {}
            module_config = helpers.read_module_config(parsed_args.module_config)
            if 'Inputs' in module_config:
                for name, path in module_config['Inputs'].items():
                    inputs[name] = path
            if 'Outputs' in module_config:
                for name, path in module_config['Outputs'].items():
                    outputs[name] = path

            configured_streams = {}
            if parsed_args.stream_configs != "unset":
                configured_streams = helpers.read_stream_configs(parsed_args.stream_configs)
            pipes_in, pipes_out = await build_network_pipes(
                inputs, outputs, configured_streams, self.node, start, end, parsed_args.force)
        else:
            # run the module within joule
            (pipes_in, pipes_out) = await helpers.build_fd_pipes(pipe_args, self.node)
        # keep track of the pipes so they can be closed
        self.pipes = list(pipes_in.values()) + list(pipes_out.values())
        return pipes_in, pipes_out

    def routes(self):
        """
        Override to register HTTP handlers for the module. Return
        an array of handlers. This creates a visualization interface.

        .. code-block:: python

            class ModuleDemo(BaseModule):

                def routes(self):
                    return [
                        web.get('/', self.index),
                        # other handlers ...
                    ]

                async def index(self, request):
                    return web.Response(text="Hello World")

                #... other module code

        """
        return []  # override in child to implement a web interface

    def _create_app(self) -> web.Application:
        routes = self.routes()
        app = web.Application()
        app.add_routes(routes)
        return app

    async def _start_interface(self, args) -> Optional[web.AppRunner]:
        routes = self.routes()
        # socket config is 'unset' when run as a standalone process
        if args.socket == 'unset':
            if len(routes) == 0:
                return None  # no routes requested, nothing to do
            # create a local server
            app = web.Application()
            app.add_routes(routes)
            runner = web.AppRunner(app)
            await runner.setup()
            site = web.TCPSite(runner, port=args.port, host=args.host)
            await site.start()
            print("starting web server at %s:%d" % (args.host, args.port))
            return runner
        # socket config is 'none' when joule does not connect a socket
        if args.socket == 'none':
            if len(routes) > 0:
                logging.error("No socket available for the interface, check module configuration")
            return None
        # otherwise start a UNIX runner on the socket
        if os.path.exists(args.socket):
            log.error("Socket address [%s] is already in use, cannot start interface" % socket)
            return None
        app = web.Application()
        app.add_routes(routes)
        runner = web.AppRunner(app)
        await runner.setup()
        sock = socket.socket(socket.AF_UNIX, socket.SOCK_STREAM)
        sock.bind(args.socket)
        site = web.SockSite(runner, sock)
        await site.start()
        print("starting web server at [%s]" % args.socket)
        return runner



import asyncio
import argparse
import logging
from typing import Dict

from joule.models.pipes.pipe import Pipe as Pipe
from joule import errors
from . import base_module


class CompositeModule(base_module.BaseModule):

    async def setup(self, parsed_args: argparse.Namespace, inputs: Dict[str, Pipe],
                    outputs: Dict[str, Pipe], loop: asyncio.AbstractEventLoop):
        """
        This method must be implemented

        Args:
            parsed_args: parsed command line arguments
            inputs: pipe connections to input streams. Keys are the names specified in the module configuration file
            outputs: pipe connections ot output streams. Keys are the names specified in the module configuration
            loop: the current event loop

        Returns:
            array of coroutine objects

        """

        assert False, "implement in child class"  # pragma: no cover

    async def run_as_task(self, parsed_args, app, loop) -> asyncio.Task:
        try:
            (pipes_in, pipes_out) = await self._build_pipes(parsed_args)
        except errors.ApiError as e:
            logging.error(str(e))
            return loop.create_task(asyncio.sleep(0))
        tasks = await self.setup(parsed_args,
                                 pipes_in,
                                 pipes_out,
                                 loop)
        return asyncio.gather(*tasks)



import numpy as np
import io
from struct import pack
import datetime
import asyncpg
from typing import List, Optional, Tuple
import logging

from joule.errors import DataError
from joule.models.stream import Stream
import joule.utilities

log = logging.getLogger('joule')

postgres_ts_offset = 946684800000000  # January 1 2000 GMT


def data_to_bytes(data: np.ndarray) -> io.BytesIO:
    pgcopy_dtype = [("num_fields", ">i2"),
                    ("time_length", '>i4'),
                    ("time", '>i8')]
    dtype_tuple = data.dtype.descr[1]
    elem_dtype = dtype_tuple[1].replace('<', '>')
    elem_length = data['data'].dtype.alignment
    if len(dtype_tuple) == 3:
        n_elem = dtype_tuple[2][0]
    else:
        n_elem = 1
    for i in range(n_elem):
        pgcopy_dtype += [("elem%d_length" % i, '>i4'),
                         ("elem%d" % i, elem_dtype)]
    pgcopy = np.empty(data.shape, pgcopy_dtype)
    pgcopy['num_fields'] = n_elem + 1
    pgcopy['time_length'] = 8
    pgcopy['time'] = data['timestamp'] - postgres_ts_offset
    if n_elem == 1:
        pgcopy['elem0_length'] = elem_length
        pgcopy['elem0'] = data['data']
    else:
        for i in range(n_elem):
            pgcopy['elem%d_length' % i] = elem_length
            pgcopy['elem%d' % i] = data['data'][:, i]
    cpy = io.BytesIO()
    # signature, flag field and and header extension (both empty)
    cpy.write(pack('!11sii', b'PGCOPY\n\377\r\n\0', 0, 0))
    cpy.write(pgcopy.tostring())
    cpy.write(pack('!h', -1))
    cpy.seek(0)
    return cpy


def bytes_to_data(buffer: io.BytesIO, dtype: np.dtype) -> np.ndarray:
    pgcopy_dtype = [("num_fields", ">i2"),
                    ("time_length", '>i4'),
                    ("time", '>i8')]
    dtype_tuple = dtype.descr[1]
    elem_dtype = dtype_tuple[1].replace('<', '>')
    elem_length = dtype['data'].alignment
    if len(dtype_tuple) == 3:
        n_elem = dtype_tuple[2][0]
    else:
        n_elem = 1
    for i in range(n_elem):
        pgcopy_dtype += [("elem%d_length" % i, '>i4'),
                         ("elem%d" % i, elem_dtype)]
    pgcopy_dtype = np.dtype(pgcopy_dtype)
    nbytes = buffer.seek(0, io.SEEK_END)
    buffer.seek(0)
    # check the header
    header = pack('!11sii', b'PGCOPY\n\377\r\n\0', 0, 0)
    rx_header = buffer.read(len(header))
    if rx_header != header:
        raise DataError("bad pgcopy header")
    row_size = pgcopy_dtype.itemsize
    if (nbytes - 21) % row_size != 0:
        raise DataError("invalid number of data bytes")
    nrows = (nbytes - 21) // row_size
    tuple_data = np.frombuffer(buffer.read(nbytes - 21), pgcopy_dtype)
    rx_data = np.empty(nrows, dtype)
    rx_data['timestamp'] = tuple_data['time'] + postgres_ts_offset
    if n_elem == 1:
        rx_data['data'] = tuple_data['elem0']
    else:
        for i in range(n_elem):
            rx_data['data'][:, i] = tuple_data['elem%d' % i]
    # reader the footer
    footer = pack('!h', -1)
    rx_footer = buffer.read()
    if footer != rx_footer:
        raise DataError("ERROR: invalid footer")
    return rx_data


def query_time_bounds(start, end):
    # bounds are [ --- )
    limits = []
    if start is not None:
        if type(start) is not datetime.datetime:
            start = datetime.datetime.fromtimestamp(start / 1e6, tz=datetime.timezone.utc)
        limits.append("time >= '%s'" % start)
    if end is not None:
        if type(end) is not datetime.datetime:
            end = datetime.datetime.fromtimestamp(end / 1e6, tz=datetime.timezone.utc)
        limits.append("time < '%s'" % end)
    if len(limits) > 0:
        return 'WHERE ' + ' AND '.join(limits)
    else:
        return ''


async def create_stream_table(conn: asyncpg.Connection, stream: Stream):
    n_elems = len(stream.elements)
    # create the main table
    col_type = get_psql_type(stream.datatype)
    cols = ["elem%d %s NOT NULL" % (x, col_type) for x in range(n_elems)]
    sql = "CREATE TABLE IF NOT EXISTS data.stream%d (" % stream.id + \
          "time TIMESTAMP NOT NULL," + \
          ', '.join(cols) + ");"
    await conn.execute(sql)
    sql = "SELECT create_hypertable('data.stream%d', 'time', if_not_exists=>true, chunk_target_size => 'estimate');" % stream.id
    await conn.execute(sql)

    # create interval table
    sql = "CREATE TABLE IF NOT EXISTS data.stream%d_intervals (" % stream.id + \
          "time TIMESTAMP NOT NULL);"
    await conn.execute(sql)


async def create_decimation_table(conn: asyncpg.Connection, stream: Stream, level: int):
    n_elems = len(stream.elements)
    table_name = 'data.stream%d_%d' % (stream.id, level)
    # create decimation table (just a template)
    mean_cols = ["elem%d REAL NOT NULL" % x for x in range(n_elems)]
    min_cols = ["elem%d_min REAL NOT NULL" % x for x in range(n_elems)]
    max_cols = ["elem%d_max REAL NOT NULL" % x for x in range(n_elems)]
    cols = mean_cols + min_cols + max_cols
    sql = "CREATE TABLE IF NOT EXISTS %s (" % table_name + \
          "time TIMESTAMP NOT NULL," + \
          ', '.join(cols) + ");"
    await conn.execute(sql)
    sql = "SELECT create_hypertable('%s', 'time', if_not_exists=>true, chunk_target_size => 'estimate');" % table_name
    await conn.execute(sql)


def get_psql_type(x: Stream.DATATYPE):
    if x == Stream.DATATYPE.FLOAT32:
        return 'real'
    elif x == Stream.DATATYPE.FLOAT64:
        return 'double precision'
    elif x == Stream.DATATYPE.INT16:
        return 'smallint'
    elif x == Stream.DATATYPE.INT32:
        return 'integer'
    elif x == Stream.DATATYPE.INT64:
        return 'bigint'
    else:
        raise DataError("Invalid type [%r] for timescale backend" % x)


async def get_row_count(conn: asyncpg.Connection, stream: Stream,
                        start=None, end=None):
    # hyper table approximate row count is not sensitive to data removal
    # always use the custom function
    try:
        bounds = await convert_time_bounds(conn, stream, start, end)
    except asyncpg.UndefinedTableError:
        return 0  # no data tables for this stream
    if bounds is None:
        return 0  # no data
    start, end = bounds
    query = "SELECT stream_row_count(%d, '%s', '%s')" % (stream.id, start, end)
    try:
        nrows = await conn.fetchval(query)
    except asyncpg.UndefinedTableError:
        return 0  # no data tables for this stream
    return nrows


async def close_interval(conn: asyncpg.Connection, stream: Stream, ts: int):
    # place a boundary 1us *after* ts
    base_table = "data.stream%d" % stream.id
    interval_table = "data.stream%d_intervals" % stream.id
    ts = datetime.datetime.fromtimestamp(ts / 1e6, tz=datetime.timezone.utc)
    # find the most recent data before this boundary (ts)
    query = "SELECT time FROM %s WHERE time <= '%s' ORDER BY time DESC LIMIT 1" % (base_table, ts)
    try:
        last_ts = await conn.fetchval(query)
    except asyncpg.UndefinedTableError:
        # no data tables so no need for an interval boundary
        return
    if last_ts is None:
        # no data exists before ts so no need for an interval boundary
        return
    # check if this interval is necessary
    query = "SELECT time FROM %s WHERE time <= '%s' ORDER BY time DESC LIMIT 1" % (interval_table, ts)
    last_interval = await conn.fetchval(query)
    if last_interval is None or last_ts > last_interval:
        query = "INSERT INTO %s(time) VALUES ($1)" % interval_table
        await conn.execute(query, last_ts + datetime.timedelta(microseconds=1))


async def get_table_names(conn: asyncpg.Connection, stream: Stream, with_schema=True) -> List[str]:
    query = r'''select table_name from information_schema.tables 
               where table_schema='data' 
               and table_type='BASE TABLE' 
               and table_name like 'stream%d\_%%';''' % stream.id
    records = await conn.fetch(query)
    if with_schema:
        return ['data.' + r['table_name'] for r in records] + ['data.stream%d' % stream.id]
    else:
        return [r['table_name'] for r in records] + ['stream%d' % stream.id]


async def get_all_table_names(conn: asyncpg.Connection, with_schema=True) -> List[str]:
    query = r'''select table_name from information_schema.tables 
               where table_schema='data' 
               and table_type='BASE TABLE' 
               and table_name like 'stream%';'''
    records = await conn.fetch(query)
    if with_schema:
        return ['data.' + r['table_name'] for r in records]
    else:
        return [r['table_name'] for r in records]


async def get_boundaries(conn: asyncpg.Connection, stream: Stream,
                         start: Optional[int], end: Optional[int]) -> List[datetime.datetime]:
    """
    Return a list of data boundaries including the start and end of the data
    """
    bounds = await convert_time_bounds(conn, stream, start, end)
    if bounds is None:
        return []  # no data so no boundaries
    start, end = bounds
    query = "SELECT time FROM data.stream%d_intervals " % stream.id
    query += query_time_bounds(start, end)
    query += " ORDER BY time ASC"
    try:
        records = await conn.fetch(query)
    except asyncpg.UndefinedTableError:
        return []  # no data tables so no boundaries
    ts = [start] + [r['time'] for r in records] + [end]
    return [x.replace(tzinfo=datetime.timezone.utc) for x in ts]


async def convert_time_bounds(conn: asyncpg.Connection,
                              stream: Stream,
                              start: Optional[int], end: Optional[int]) -> Optional[
    Tuple[datetime.datetime, datetime.datetime]]:
    """
    Convert Unix us timestamps to datetime objects and populate [None] values with the
    start or end of the data respectively
    """
    x = start
    y = end
    if start is None:
        query = "SELECT time FROM data.stream%d ORDER BY time ASC LIMIT 1" % stream.id
        try:
            start = await conn.fetchval(query)
        except asyncpg.UndefinedTableError:
            return None  # no data tables so no valid time bounds
        if start is None:
            # remove intervals?
            return None  # no data so no valid time bounds
        start = start.replace(tzinfo=datetime.timezone.utc)

    else:
        start = datetime.datetime.fromtimestamp(start / 1e6, tz=datetime.timezone.utc)
    if end is None:
        query = "SELECT time FROM data.stream%d ORDER BY time DESC LIMIT 1" % stream.id
        try:
            end = await conn.fetchval(query) + datetime.timedelta(microseconds=1)
        except asyncpg.UndefinedTableError:
            return None  # no data tables so no valid time bounds
        if end is None:
            # remove intervals?
            return None  # no data so no valid time bounds
        end = end.replace(tzinfo=datetime.timezone.utc)
    else:
        end = datetime.datetime.fromtimestamp(end / 1e6, tz=datetime.timezone.utc)
    return start, end


async def limit_time_bounds(conn: asyncpg.Connection,
                            stream: Stream,
                            start: Optional[int], end: Optional[int]) -> Tuple[Optional[int], Optional[int]]:
    data_bounds = await convert_time_bounds(conn, stream, None, None)
    if data_bounds is None:
        return None, None
    data_start = joule.utilities.datetime_to_timestamp(data_bounds[0])
    data_end = joule.utilities.datetime_to_timestamp(data_bounds[1])
    if start is None:
        start = data_start
    else:
        start = max(start, data_start)
    if end is None:
        end = data_end
    else:
        end = min(end, data_end)
    return start, end



import asyncio
import aiohttp
import numpy as np
import random
import time
from typing import List, Callable
import logging

from joule.models import Stream, pipes
from joule.models.data_store import errors
from joule.models.data_store.nilmdb_helpers import compute_path, ERRORS, check_for_error

Loop = asyncio.AbstractEventLoop
log = logging.getLogger('joule')


class Inserter:

    def __init__(self, server: str, stream: Stream, insert_period: float, cleanup_period: float,
                 session_factory: Callable[[], aiohttp.ClientSession], retry_interval=0.5):
        self.insert_url = "{server}/stream/insert".format(server=server)
        self.remove_url = "{server}/stream/remove".format(server=server)
        self.create_url = "{server}/stream/create".format(server=server)

        self.server = server  # save for initializing decimators
        self.stream = stream
        self.path = compute_path(stream)
        # add offsets to the period to distribute traffic
        self.insert_period = insert_period + insert_period * random.random() * 0.5
        self.cleanup_period = cleanup_period + cleanup_period * random.random() * 0.25
        self._get_client = session_factory
        self.decimator = None
        self.retry_interval = retry_interval

    async def run(self, pipe: pipes.Pipe, loop: Loop) -> None:
        """insert stream data from the queue until the queue is empty"""
        # create the database path
        # lazy stream creation,
        try:
            await self._create_path()
        except asyncio.CancelledError:
            return

        cleaner_task: asyncio.Task = None
        if self.stream.keep_us != Stream.KEEP_ALL:
            cleaner_task = loop.create_task(self._clean())

        while True:
            try:
                async with self._get_client() as session:
                    last_ts = None
                    while True:
                        await asyncio.sleep(self.insert_period)
                        data = await pipe.read()
                        # there might be an interval break and no new data
                        if len(data) > 0:
                            if last_ts is not None:
                                start = last_ts
                            else:
                                start = data['timestamp'][0]
                            end = data['timestamp'][-1] + 1
                            last_ts = end
                            # lazy initialization of decimator
                            if self.stream.decimate and self.decimator is None:
                                self.decimator = NilmdbDecimator(self.server, self.stream, 1, 4,
                                                                 self._get_client)
                            # send the data
                            params = {"start": "%d" % start,
                                      "end": "%d" % end,
                                      "path": self.path,
                                      "binary": '1'}
                            async with session.put(self.insert_url, params=params,
                                                   data=data.tostring()) as resp:
                                if resp.status != 200:
                                    error = await resp.text()
                                    if cleaner_task is not None:
                                        cleaner_task.cancel()
                                        await cleaner_task
                                    raise errors.DataError("NilmDB error: %s" % error)
                            # this was successful so consume the data
                            pipe.consume(len(data))
                            # decimate the data
                            if self.decimator is not None:
                                await self.decimator.process(data)
                        # check for interval breaks
                        if pipe.end_of_interval:
                            last_ts = None
                            if self.decimator is not None:
                                self.decimator.close_interval()
            except aiohttp.ClientError as e:  # pragma: no cover
                log.warning("NilmDB raw inserter error: %r, retrying request" % e)
                await asyncio.sleep(self.retry_interval)  # retry the request
            except (pipes.EmptyPipe, asyncio.CancelledError):
                break  # terminate the inserter
        if cleaner_task is not None:
            cleaner_task.cancel()
            await cleaner_task

    async def _create_path(self):
        data = {"path": compute_path(self.stream),
                "layout": self.stream.layout}
        while True:
            try:
                async with self._get_client() as session:
                    async with session.post(self.create_url, data=data) as resp:
                        await check_for_error(resp, ignore=[ERRORS.STREAM_ALREADY_EXISTS])
                        break
            except aiohttp.ClientError as e:  # pragma: no cover
                log.warning("NilmDB inserter create_path error: %r, retrying request" % e)
                await asyncio.sleep(self.retry_interval)  # retry the request

    async def _clean(self):
        while True:
            try:
                async with self._get_client() as session:
                    while True:
                        await asyncio.sleep(self.cleanup_period)
                        keep_time = int(time.time() * 1e6) - self.stream.keep_us
                        # remove raw data
                        params = {"start": "%d" % 0,
                                  "end": "%d" % keep_time,
                                  "path": self.path}
                        async with session.post(self.remove_url, params=params) as resp:
                            if resp.status != 200:  # pragma: no cover
                                msg = await resp.text()
                                log.error("NilmDB cleaning error: %s" % msg)
                        # remove decimation data
                        if self.decimator is not None:
                            for path in self.decimator.get_paths():
                                params["path"] = path
                                async with session.post(self.remove_url, params=params) as resp:
                                    if resp.status != 200:  # pragma: no cover
                                        msg = await resp.text()
                                        log.error("NilmDB cleaning error: %s" % msg)
            except aiohttp.ClientError as e:  # pragma: no cover
                log.warning("NilmDB cleaning error: %r" % e)
            except asyncio.CancelledError:
                break


class NilmdbDecimator:

    def __init__(self, server: str, stream: Stream, from_level: int, factor: int,
                 session_factory: Callable[[], aiohttp.ClientSession], retry_interval=0.5):
        self.stream = stream
        self.level = from_level * factor
        self.insert_url = "{server}/stream/insert".format(server=server)
        self.create_url = "{server}/stream/create".format(server=server)
        self.server = server
        self.path = compute_path(stream, self.level)
        if from_level > 1:
            self.again = True
        else:
            self.again = False
        self.factor = factor
        self.layout = stream.decimated_layout
        self.buffer = []
        self.last_ts = None
        self.path_created = False
        self.child: NilmdbDecimator = None
        self.retry_interval = retry_interval
        self._get_client = session_factory
        # hold off to rate limit NilmDB traffic
        self.holdoff = 0  # random.random()

    async def process(self, data: np.ndarray) -> None:
        """decimate data and insert it, retry on error"""
        while True:
            try:
                if not self.path_created:
                    await self._create_path()
                    self.path_created = True

                async with self._get_client() as session:
                    decim_data = self._process(data)
                    if len(decim_data) == 0:
                        return
                    if self.last_ts is not None:
                        start = self.last_ts
                    else:
                        start = decim_data['timestamp'][0]
                    end = decim_data['timestamp'][-1] + 1
                    self.last_ts = end
                    # lazy initialization of child
                    if self.child is None:
                        self.child = NilmdbDecimator(self.server, self.stream, self.level,
                                                     self.factor, self._get_client)
                    params = {"start": "%d" % start,
                              "end": "%d" % end,
                              "path": self.path,
                              "binary": '1'}
                    async with session.put(self.insert_url, params=params,
                                           data=decim_data.tostring()) as resp:
                        if resp.status != 200:  # pragma: no cover
                            error = await resp.text()
                            raise errors.DataError("NilmDB(d) error: %s" % error)
                    # feed data to child decimator
                    await self.child.process(decim_data)
                    await asyncio.sleep(self.holdoff)
                    break  # success, leave the loop
            except aiohttp.ClientError as e:  # pragma: no cover
                log.warning("NilmDB decimation error: %r, retrying request" % e)
                await asyncio.sleep(self.retry_interval)  # retry the request
            except asyncio.CancelledError:  # pragma: no cover
                break

    def close_interval(self):
        self.buffer = []
        self.last_ts = None
        if self.child is not None:
            self.child.close_interval()

    def get_paths(self) -> List[str]:
        paths = [compute_path(self.stream, self.level)]
        if self.child is not None:
            paths = paths + self.child.get_paths()
        return paths

    def _process(self, sarray: np.ndarray) -> np.ndarray:

        # flatten structured array
        data = np.c_[sarray['timestamp'][:, None], sarray['data']]

        # check if there is old data
        if len(self.buffer) != 0:
            # append the new data onto the old data
            data = np.concatenate((self.buffer, data))
        (n, m) = data.shape

        # Figure out which columns to use as the input for mean, min, and max,
        # depending on whether this is the first decimation or we're decimating
        # again.  Note that we include the timestamp in the means.
        if self.again:
            c = (m - 1) // 3
            # e.g. c = 3
            # ts mean1 mean2 mean3 min1 min2 min3 max1 max2 max3
            mean_col = slice(0, c + 1)
            min_col = slice(c + 1, 2 * c + 1)
            max_col = slice(2 * c + 1, 3 * c + 1)
        else:
            mean_col = slice(0, m)
            min_col = slice(1, m)
            max_col = slice(1, m)

        # Discard extra rows that aren't a multiple of factor
        n = n // self.factor * self.factor

        if n == 0:  # not enough data to work with, save it for later
            self.buffer = data
            return np.array([])

        trunc_data = data[:n, :]
        # keep the leftover data
        self.buffer = np.copy(data[n:, :])

        # Reshape it into 3D so we can process 'factor' rows at a time
        trunc_data = trunc_data.reshape(n // self.factor, self.factor, m)

        # Fill the result
        out = np.c_[np.mean(trunc_data[:, :, mean_col], axis=1),
                    np.min(trunc_data[:, :, min_col], axis=1),
                    np.max(trunc_data[:, :, max_col], axis=1)]

        # structure the array
        width = np.shape(out)[1] - 1
        dtype = np.dtype([('timestamp', '<i8'), ('data', '<f4', width)])
        sout = np.zeros(out.shape[0], dtype=dtype)
        sout['timestamp'] = out[:, 0]
        sout['data'] = out[:, 1:]
        # insert the data into the database
        return sout

    async def _create_path(self):
        data = {"path": compute_path(self.stream, self.level),
                "layout": self.stream.decimated_layout}
        while True:
            try:
                async with self._get_client() as session:
                    async with session.post(self.create_url, data=data) as resp:
                        await check_for_error(resp, ignore=[ERRORS.STREAM_ALREADY_EXISTS])
                        break
            except aiohttp.ClientError as e:  # pragma: no cover
                log.warning("NilmDB decimator create_path error: %r, retrying request" % e)
                await asyncio.sleep(self.retry_interval)  # retry the request



import aiohttp
from joule.models import Stream
from enum import Enum
from typing import List

from joule.models.data_store import errors


class ERRORS(Enum):
    STREAM_ALREADY_EXISTS = "stream already exists at this path"
    NO_SUCH_STREAM = "No such stream"
    NO_STREAM_AT_PATH = "No stream at path"


def compute_path(stream: Stream, decimation_level: int = 1):
    path = "/joule/%d" % stream.id
    if decimation_level == 1:
        return path
    return path + "~decim-%d" % decimation_level


async def check_for_error(resp: aiohttp.ClientResponse, ignore: List[ERRORS] = None):
    if resp.status == 200:
        return  # OK
    try:
        error = await resp.json()
        if ignore is not None:
            for error_type in ignore:
                if error_type.value in error["message"]:
                    return  # OK
    except aiohttp.ContentTypeError:
        raise errors.DataError("[%d] invalid json response \"%s\"" % (resp.status, await resp.text()))
    raise errors.DataError("[%d]: %s" % (resp.status, error["message"]))



from joule.client import FilterModule
import textwrap
import numpy as np
from typing import Dict, List
from joule import Pipe
from joule.errors import ApiError
from scipy.interpolate import interp1d

ARGS_DESC = """
---
:name:
  Merge Filter
:author:
  John Donnal
:license:
  Open
:url:
  http://git.wattsworth.net/wattsworth/joule.git
:description:
  combine multiple input streams into a single output stream
:usage:
  Specify a master stream and one or more additional streams.
  The output will have timestamps from the master stream and linearly
  interpolated data from all the additional streams
  
  
    | Arguments | Description        |
    |-----------|--------------------|
    |``master`` | name of master stream |

  ```
:inputs:
  master
  : N elements
  additional
  : M elements

:outputs:
  output
  :  float32 with N+M elements auto detected from input source

:stream_configs:
  #output#
     [Main]
     name = Merged
     path = /path/to/output
     datatype = float32
     keep = 1w

     # [width] number of elements of all input streams
     [Element1]
     name = Data Column 1

     [Element2]
     name = Data Column 2

     #additional elements...

:module_config:
    [Main]
    name = Merge Filter
    exec_cmd = joule-merge-filter

    [Arguments]
    master = input1 

    [Inputs]
    input1 = /path/to/input1
    input2 = /path/to/input2
    #... additional inputs as desired
    
    [Outputs]
    output = /path/to/output
---
"""


class MergeFilter(FilterModule):
    """Merge input streams"""

    def custom_args(self, parser):  # pragma: no cover
        grp = parser.add_argument_group("module",
                                        "module specific arguments")
        grp.add_argument("--master", type=str, required=True,
                         help="name of master input stream")
        parser.description = textwrap.dedent(ARGS_DESC)

    async def run(self, parsed_args, inputs: Dict[str, Pipe], outputs):
        master_name = parsed_args.master

        # ---- VERIFY PARAMETERS and STREAM FORMATS -----
        if master_name not in inputs:
            raise Exception("the specified master [%s] is not in the inputs" % parsed_args.master)
        master = inputs[master_name]
        slaves = [inputs[name] for name in inputs.keys() if name != master_name]
        if len(outputs) != 1:
            raise Exception("must specify one output stream")
        output = list(outputs.values())[0]
        input_width = np.sum([slave.width for slave in slaves]) + master.width
        if input_width != output.width:
            raise Exception("output must have %d elements" % input_width)
        # ---- END VERIFY ------

        if not await align_streams(master, slaves):
            return
        while not self.stop_requested:
            try:
                master_data = await master.read()
                # allocate the output array (we won't have any *more* data than this)
                output_data = np.empty(len(master_data), dtype=output.dtype)
                # put in the master data and the timestamps
                output_data['timestamp'] = master_data['timestamp']
                output_data['data'][:, :master.width] = make2d(master_data['data'])
                valid_rows = len(output_data)
                offset = master.width
                for slave in slaves:
                    valid_rows = min(await get_data(slave, output_data, offset),
                                     valid_rows)
                    offset += slave.width
                # consume the right amount from each slave based on the last valid timestamp
                for slave in slaves:
                    await consume_data(slave, output_data['timestamp'][valid_rows - 1])

                # write the output
                await output.write(output_data[:valid_rows])

                # consume the used data and check for interval breaks
                interval_break = False
                master.consume(valid_rows)

                # stay within the same interval until all streams catch up
                if master.end_of_interval and len(master_data) > valid_rows:
                    master.reread_last()
                    master.interval_break = False

                if master.end_of_interval:
                    interval_break = True
                for slave in slaves:
                    if slave.end_of_interval:
                        interval_break = True
                if interval_break:
                    await output.close_interval()
                    if not await align_streams(master, slaves):
                        break


            except ApiError:
                break
        await output.close()


def make2d(arr: np.ndarray) -> np.ndarray:
    if len(arr.shape) == 1:
        return arr[:, None]
    else:
        return arr


async def align_streams(master: Pipe, slaves: List[Pipe]):
    # we need a previous sample of every slave
    try:
        slave_datas = [await slave.read() for slave in slaves]
        master_data = await master.read()

        # get the maximum slave start time (we can't interpolate before this)
        latest_slave_start = np.max([data['timestamp'][0] for data in slave_datas])
        # make sure all streams have some data larger than the latest_slave_start
        while master_data['timestamp'][-1] < latest_slave_start:
            master.consume(len(master_data))
            master_data = await master.read()
        for i in range(len(slaves)):
            while slave_datas[i]['timestamp'][-1] < latest_slave_start:
                slaves[i].consume(len(slave_datas[i]))
                slave_datas[i] = await slaves[i].read()
        # find the closest ts in master that is larger than latest_slave_start
        start_ts = master_data['timestamp'][master_data['timestamp'] >= latest_slave_start][0]
        i = 0
        for data in slave_datas:
            # find the closest ts in slaves that is smaller than latest_slave_start
            idx = np.argmax(data['timestamp'][data['timestamp'] <= start_ts])
            # consume up to this point
            slaves[i].consume(idx)
            i += 1
        # consume the master before start_ts
        too_early = master_data['timestamp'][master_data['timestamp'] < start_ts]
        master.consume(len(too_early))

        # to preserve any interval breaks, reread the same data again
        master.reread_last()
        for slave in slaves:
            slave.reread_last()
    except ApiError:
        return False
    return True


async def get_data(source_pipe: Pipe, dest: np.ndarray, offset: int) -> float:
    source = await source_pipe.read()
    ts = dest['timestamp']
    max_ts = source['timestamp'][-1]
    f = interp1d(source['timestamp'], source['data'], axis=0, fill_value='extrapolate')
    resampled_data = f(ts[ts <= max_ts])
    dest['data'][:len(resampled_data), offset:offset + source_pipe.width] = make2d(resampled_data)
    last_valid_ts = ts[len(resampled_data) - 1]
    dest_rows = np.argwhere(ts == last_valid_ts)[0][0] + 1
    return dest_rows


async def consume_data(source_pipe: Pipe, ts: int) -> None:
    # return
    source_pipe.reread_last()
    source = await source_pipe.read()
    used_rows = len(source['timestamp'][source['timestamp'] <= ts])
    source_pipe.consume(used_rows)
    # stay within the same interval until all streams catch up
    if source_pipe.end_of_interval and used_rows < len(source):
        source_pipe.reread_last()
        source_pipe.interval_break = False


def main():  # pragma: no cover
    r = MergeFilter()
    r.start()


if __name__ == "__main__":
    main()



import argparse
import sys
import os
import configparser
import dateparser
import re
from typing import Tuple, Dict

from joule.errors import ConfigurationError
from joule.models import stream
from joule.api import stream as api_stream
from joule.services.helpers import load_configs

""" helpers for handling module arguments
    Complicated because we want to look for module_config
    and add these args but this should be transparent
    (eg, the -h flag should show the real help)
"""


def module_args():
    # build a dummy parser to look for module_config
    temp_parser = argparse.ArgumentParser()
    temp_parser.add_argument("--module_config", default="unset")
    arg_list = sys.argv[1:]  # ignore the program name
    stdout = sys.stdout
    sys.stdout = open(os.devnull, 'w')
    try:
        args = temp_parser.parse_known_args()[0]
        if args.module_config != "unset":
            arg_list += _append_args(args.module_config)
    except SystemExit:
        pass
    finally:
        sys.stdout.close()
        sys.stdout = stdout
    return arg_list


def load_args_from_file(path):
    args = ["--module_config", path]
    return args + _append_args(path)


def _append_args(module_config_file):
    from joule.models import module
    # if a module_config is specified add its args
    module_config = configparser.ConfigParser()
    with open(module_config_file, 'r') as f:
        module_config.read_file(f)
    my_module = module.from_config(module_config)
    args = []
    for (arg, value) in my_module.arguments.items():
        args += ["--" + arg, value]
    return args


def read_module_config(file_path: str) -> configparser.ConfigParser:
    config = configparser.ConfigParser()
    try:
        # can only happen if process does not have read permissions
        if len(config.read(file_path)) != 1:
            raise ConfigurationError("Cannot read module config [%s]" % file_path)
    except configparser.Error as e:
        raise ConfigurationError("Invalid module config: %s" % str(e))
    return config


def read_stream_configs(file_path: str) -> Dict[str, stream.Stream]:
    configs = load_configs(file_path)
    streams: Dict[str, stream.Stream] = {}
    for file_path, data in configs.items():
        try:
            s = stream.from_config(data)
            # convert to an API object
            api_s = api_stream.from_json(s.to_json())
            stream_path = _create_path(data)

            streams[stream_path] = api_s
        except KeyError as e:
            raise ConfigurationError("Invalid stream [%s]: [Main] missing %s" %
                                     (file_path, e.args[0]))
    return streams


def validate_time_bounds(start: str, end: str) -> Tuple[int, int]:
    if start is not None:
        start = int(dateparser.parse(start).timestamp() * 1e6)
    if end is not None:
        end = int(dateparser.parse(end).timestamp() * 1e6)
    if start is not None and end is not None:
        if start >= end:
            raise ConfigurationError("start [%d] must be < end [%d]" % (start, end))
    return start, end


def _create_path(data: configparser.ConfigParser) -> str:
    #
    path = data['Main']['path']
    if path != '/' and re.fullmatch(r'^(/[\w -]+)+$', path) is None:
        raise ConfigurationError(
            "invalid path, use format: /dir/subdir/../file. "
            "valid characters: [0-9,A-Z,a-z,_,-, ]")

    return '/'.join([path, data['Main']['name']])



import enum
from typing import Dict
import configparser

from joule.errors import ConfigurationError
from joule.models.stream import Stream

"""
Configuration File:
[Main]
name = module name
description = module description
exec_cmd = /path/to/file
is_app = no

[Arguments]
key = value

[Inputs]
#name = full_path:<stream config>
labjack = /labjack/device3/data:float32[e0,e1,e2]
path2 = /nilmdb/input/stream2
  ....
pathN = /nilmdb/input/streamN

[Outputs]
#name = full_path:<stream config>
path1 = /nilmdb/output/stream1:float32[x,y,z]
path2 = /nilmdb/output/stream2
  ....
pathN = /nilmdb/output/streamN
"""


class Module:
    class STATUS(enum.Enum):
        LOADED = enum.auto()
        RUNNING = enum.auto()
        FAILED = enum.auto()
        UNKNOWN = enum.auto()

    def __init__(self, name: str, exec_cmd: str, description: str = "",
                 log_size: int = 100,
                 is_app: bool = False, uuid: int = None):
        self.name = name
        self.exec_cmd = exec_cmd
        self.description = description
        self.is_app = is_app
        self.log_size = log_size
        # arg_name => value
        self.arguments: Dict[str, str] = {}
        # pipe name (from config) => stream object
        self.inputs: Dict[str, Stream] = {}
        self.outputs: Dict[str, Stream] = {}
        self.uuid: int = uuid
        self.status: Module.STATUS = Module.STATUS.UNKNOWN

    def __repr__(self):
        return "<Module(uuid=%r, name=%s)>" % (self.uuid, self.name)

    def to_json(self):
        return {
            'id': self.uuid,
            'name': self.name,
            'description': self.description,
            'exec_cmd': self.exec_cmd,
            'is_app': self.is_app,
            'arguments': self.arguments,
            'inputs': dict((name, stream.id) for (name, stream) in self.inputs.items()),
            'outputs': dict((name, stream.id) for (name, stream) in self.outputs.items())
        }


def from_config(config: configparser.ConfigParser) -> Module:
    try:
        main_configs: configparser.ConfigParser = config["Main"]
    except KeyError as e:
        raise ConfigurationError("Missing section [%s]" % e.args[0]) from e
    try:
        name = validate_name(main_configs["name"])
        description = main_configs.get("description", fallback="")
        exec_cmd = main_configs["exec_cmd"]
        is_app = main_configs.getboolean("is_app", fallback=False)
    except KeyError as e:
        raise ConfigurationError("[Main] missing %s" % e.args[0]) from e
    my_module = Module(name=name, description=description, exec_cmd=exec_cmd,
                       is_app=is_app)
    # parse the arguments
    if 'Arguments' in config:
        for name, value in config['Arguments'].items():
            my_module.arguments[name] = value
    return my_module


def validate_name(name: str) -> str:
    if len(name) == 0:
        raise ConfigurationError("missing name")
    return name



import click
from typing import Dict, List, Tuple, Optional, Union
from operator import attrgetter
import aiohttp
import asyncio
import json
import numpy as np
import datetime

from joule.cli.config import pass_config
from joule.api import get_node, BaseNode, Annotation
from joule.api.annotation import from_json as annotation_from_json
from joule.models import stream, Stream, pipes, StreamInfo
from joule.utilities import interval_difference, human_to_timestamp, timestamp_to_human
from joule import errors

Interval = Tuple[int, int]


@click.command(name="copy")
@click.option('-s', "--start", help="timestamp or descriptive string")
@click.option('-e', "--end", help="timestamp or descriptive string")
@click.option('-n', '--new', help="copy starts at the last timestamp of the destination", is_flag=True)
@click.option('-d', '--destination-node', help="node name or Nilmdb URL")
@click.option('--source-url', help="copy from a Nilmdb URL")
@click.argument("source")
@click.argument("destination")
@pass_config
def data_copy(config, start, end, new, destination_node, source_url, source, destination):
    loop = asyncio.get_event_loop()
    try:
        loop.run_until_complete(_run(config, start, end, new, destination_node,
                                     source_url, source, destination))
    except errors.ApiError as e:
        raise click.ClickException(str(e)) from e
    finally:
        loop.run_until_complete(
            config.close_node())
        loop.close()


async def _run(config, start, end, new, destination_node, source_url, source, destination):
    # determine if the source node is NilmDB or Joule
    if source_url is None:
        source_node = config.node
        nilmdb_source = False
    else:
        source_node = source_url
        await _validate_nilmdb_url(source_node)
        nilmdb_source = True

    # determine if the destination node is NilmDB or Joule
    nilmdb_dest = False

    try:
        if destination_node is None:
            dest_node = config.node
        else:
            dest_node = get_node(destination_node)
    except errors.ApiError:
        nilmdb_dest = True
        dest_node = destination_node
        await _validate_nilmdb_url(dest_node)

    # retrieve the source stream
    src_stream = await _retrieve_source(source_node, source, is_nilmdb=nilmdb_source)

    # retrieve the destination stream (create it if necessary)
    dest_stream = await _retrieve_destination(dest_node, destination, src_stream, is_nilmdb=nilmdb_dest)
    # make sure streams are compatible
    if src_stream.layout != dest_stream.layout:
        raise errors.ApiError("Error: source (%s) and destination (%s) datatypes are not compatible" % (
            src_stream.layout, dest_stream.layout))
    # warn if the elements are not the same
    element_warning = False
    src_elements = sorted(src_stream.elements, key=attrgetter('index'))
    dest_elements = sorted(dest_stream.elements, key=attrgetter('index'))
    for i in range(len(src_elements)):
        if src_elements[i].name != dest_elements[i].name:
            element_warning = True
        if src_elements[i].units != dest_elements[i].units:
            element_warning = True
    if (element_warning and
            not click.confirm("WARNING: Element configurations do not match. Continue?")):
        click.echo("Cancelled")
        return
    # if new is set start and end may not be specified
    if new and start is not None:
        raise click.ClickException("Error: either specify 'new' or a starting timestamp, not both")

    # make sure the time bounds make sense
    if start is not None:
        try:
            start = human_to_timestamp(start)
        except ValueError:
            raise errors.ApiError("invalid start time: [%s]" % start)
    if end is not None:
        try:
            end = human_to_timestamp(end)
        except ValueError:
            raise errors.ApiError("invalid end time: [%s]" % end)
    if (start is not None) and (end is not None) and ((end - start) <= 0):
        raise click.ClickException("Error: start [%s] must be before end [%s]" % (
            datetime.datetime.fromtimestamp(start / 1e6),
            datetime.datetime.fromtimestamp(end / 1e6)))
    if new:
        # pull all the destination intervals and use the end of the last one as the 'end' for the copy
        dest_intervals = await _get_intervals(dest_node, dest_stream, destination, None, None, is_nilmdb=nilmdb_dest)
        if len(dest_intervals) > 0:
            start = dest_intervals[-1][-1]
        print("Starting copy at [%s]" % timestamp_to_human(start))
    # compute the target intervals (source - dest)
    src_intervals = await _get_intervals(source_node, src_stream, source, start, end, is_nilmdb=nilmdb_source)
    dest_intervals = await _get_intervals(dest_node, dest_stream, destination, start, end, is_nilmdb=nilmdb_dest)
    new_intervals = interval_difference(src_intervals, dest_intervals)
    existing_intervals = interval_difference(src_intervals, new_intervals)

    async def _copy(intervals):
        # compute the duration of data to copy
        duration = 0
        for interval in intervals:
            duration += interval[1] - interval[0]

        with click.progressbar(
                label='Copying data',
                length=duration) as bar:
            for interval in intervals:
                await _copy_interval(interval[0], interval[1], bar)
                await _copy_annotations(interval[0], interval[1])

    async def _copy_annotations(istart, iend):
        if nilmdb_source:
            src_annotations = await _get_nilmdb_annotations(source_node, source, istart, iend)
        else:
            src_annotations = await source_node.annotation_get(src_stream.id, start=istart, end=iend)

        if nilmdb_dest:
            dest_annotations = await _get_nilmdb_annotations(dest_node, destination, istart, iend)
            new_annotations = [a for a in src_annotations if a not in dest_annotations]
            if len(new_annotations) > 0:
                # create ID's for the new annotations
                if len(dest_annotations) > 0:
                    id_val = max([a.id for a in dest_annotations]) + 1
                else:
                    id_val = 0
                for a in new_annotations:
                    a.id = id_val
                    id_val += 1
                await _create_nilmdb_annotations(dest_node, destination, new_annotations + dest_annotations)
        else:
            dest_annotations = await dest_node.annotation_get(dest_stream.id, start=istart, end=iend)
            new_annotations = [a for a in src_annotations if a not in dest_annotations]
            for annotation in new_annotations:
                await dest_node.annotation_create(annotation, dest_stream.id)

    if len(new_intervals) == 0:
        if len(src_intervals) > 0:
            click.echo("Nothing to copy, syncing annotations")
            for interval in src_intervals:
                await _copy_annotations(interval[0], interval[1])
        else:
            click.echo("Nothing to copy")
        # clean up
        if not nilmdb_dest:
            await dest_node.close()
        if not nilmdb_source:
            await source_node.close()
        return

    async def _copy_interval(istart, iend, bar):
        #print("[%s] -> [%s]" % (timestamp_to_human(istart), timestamp_to_human(iend)))
        if nilmdb_source:
            src_params = {'path': source, 'binary': 1,
                          'start': istart, 'end': iend}
            src_url = "{server}/stream/extract".format(server=source_node)
            src_headers = {}
            src_ssl = None
        else:
            src_params = {'id': src_stream.id, 'start': istart, 'end': iend}
            src_url = "{server}/data".format(server=source_node.session.url)
            src_headers = {"X-API-KEY": source_node.session.key}
            src_ssl = source_node.session.ssl_context
        async with aiohttp.ClientSession(
                timeout=aiohttp.ClientTimeout(total=None)) as session:
            async with session.get(src_url,
                                   params=src_params,
                                   headers=src_headers,
                                   ssl=src_ssl) as src_response:
                if src_response.status != 200:
                    msg = await src_response.text()
                    if msg == 'this stream has no data':
                        # This is not an error because a previous copy may have been interrupted
                        # This will cause the destination to have an interval gap where the source has no data
                        # Example:   source:  |**     *******|
                        #            dest:    |** |  |*******|
                        #                          ^--- looks like missing data but there's nothing in the source
                        return  # ignore empty intervals
                    raise click.ClickException("Error reading from source: %s" % msg)

                pipe = pipes.InputPipe(stream=dest_stream, reader=src_response.content)

                async def _data_sender():

                    last_ts = istart
                    try:
                        while True:
                            data = await pipe.read()
                            pipe.consume(len(data))
                            if len(data) > 0:
                                cur_ts = data[-1]['timestamp']
                                yield data.tostring()
                                # total time extents of this chunk
                                bar.update(cur_ts - last_ts)
                                last_ts = cur_ts
                            # if pipe.end_of_interval:
                            #    yield pipes.interval_token(dest_stream.layout). \
                            #        tostring()
                    except pipes.EmptyPipe:
                        pass
                    bar.update(iend - last_ts)

                if nilmdb_dest:
                    dst_params = {"start": istart, "end": iend, "path": destination, "binary": 1}
                    dst_url = "{server}/stream/insert".format(server=dest_node)
                    await _send_nilmdb_data(dst_url,
                                            dst_params, _data_sender(),
                                            pipes.compute_dtype(dest_stream.layout),
                                            session)
                else:
                    dst_url = "{server}/data".format(server=dest_node.session.url)
                    dst_params = {"id": dest_stream.id}
                    dst_headers = {"X-API-KEY": dest_node.session.key}
                    dst_ssl = dest_node.session.ssl_context
                    async with session.post(dst_url,
                                            params=dst_params,
                                            data=_data_sender(),
                                            headers=dst_headers,
                                            ssl=dst_ssl,
                                            chunked=True) as dest_response:
                        if dest_response.status != 200:
                            msg = await dest_response.text()
                            raise errors.ApiError("Error writing to destination: %s" % msg)

    try:
        # copy over any new annotations from existing intervals
        for interval in existing_intervals:
            await _copy_annotations(interval[0], interval[1])
        await _copy(new_intervals)
        click.echo("OK")
    # this should be caught by the stream info requests
    # it is only generated if the joule server stops during the
    # data read/write
    except aiohttp.ClientError as e:  # pragma: no cover
        raise click.ClickException("Error: %s" % str(e))
    finally:
        if not nilmdb_dest:
            await dest_node.close()
        if not nilmdb_source:
            await source_node.close()


async def _get_intervals(server: Union[BaseNode, str], my_stream: Stream, path: str,
                         start: Optional[int], end: Optional[int],
                         is_nilmdb: bool = False) -> List[Interval]:
    if is_nilmdb:
        intervals = []
        params = {"path": path}
        if start is not None:
            params['start'] = start
        if end is not None:
            params['end'] = end
        url = "{server}/stream/intervals".format(server=server)
        async with aiohttp.ClientSession() as session:
            async with session.get(url, params=params) as resp:
                if not resp.status == 200:
                    raise errors.ApiError("unable to retrieve intervals for [%s]" % path)
                body = await resp.text()
                if body == '':
                    return intervals
                for line in body.strip().split("\n"):
                    intervals.append(json.loads(line))
        return intervals

    else:
        return await server.data_intervals(my_stream, start, end)


async def _retrieve_source(server: Union[BaseNode, str], path: str, is_nilmdb: bool = False) -> Stream:
    if is_nilmdb:
        src_stream, src_info = await _retrieve_nilmdb_stream(server, path)
        if src_stream is None:
            raise errors.ApiError("The stream [%s] is not available on [%s]" % (path, server))
    else:
        src_stream = await server.stream_get(path)
        src_info = await server.stream_info(path)

    if src_info.start is None or src_info.end is None:
        raise errors.ApiError("[%s] has no data" % path)
    return src_stream


async def _retrieve_destination(server: Union[str, BaseNode], path: str, template: Stream,
                                is_nilmdb: bool = False) -> Stream:
    if is_nilmdb:
        dest_stream, dest_info = await _retrieve_nilmdb_stream(server, path)
        if dest_stream is None:
            return await _create_nilmdb_stream(server, path, template)
        else:
            return dest_stream
    else:
        try:
            return await server.stream_get(path)
        except errors.ApiError as e:
            if "404" in str(e):
                return await _create_joule_stream(server, path, template)
            raise e


async def _create_joule_stream(node: BaseNode, path: str, template: Stream) -> Stream:
    click.echo("creating destination stream")
    # split the destination into the path and stream name
    dest_stream = stream.from_json(template.to_json())
    dest_stream.keep_us = Stream.KEEP_ALL
    dest_stream.name = path.split("/")[-1]
    folder = "/".join(path.split("/")[:-1])
    return await node.stream_create(dest_stream, folder)


async def _create_nilmdb_stream(server: str, path: str, template: Stream):
    click.echo("creating destination stream")
    dest_stream = stream.from_json(template.to_json())
    dest_stream.keep_us = Stream.KEEP_ALL
    dest_stream.name = path.split("/")[-1]

    url = "{server}/stream/create".format(server=server)
    data = {"path": path,
            "layout": dest_stream.layout}
    async with aiohttp.ClientSession() as session:
        async with session.post(url, data=data) as resp:
            if not resp.status == 200:
                raise click.ClickException("cannot create [%s] on [%s]" % (path, server))

        # add the metadata
        url = "{server}/stream/set_metadata".format(server=server)
        data = {
            "path": path,
            "data": json.dumps({"config_key__": json.dumps(dest_stream.to_nilmdb_metadata())})
        }
        async with session.post(url, data=data) as resp:
            if not resp.status == 200:
                raise click.ClickException("cannot create metadata for [%s] on [%s]" % (path, server))
    return dest_stream


async def _retrieve_nilmdb_stream(server: str, path: str) -> Tuple[Optional[Stream], Optional[StreamInfo]]:
    url = "{server}/stream/get_metadata".format(server=server)
    params = {"path": path, "key": 'config_key__'}
    async with aiohttp.ClientSession() as session:
        async with session.get(url, params=params) as resp:
            if resp.status == 404:
                return None, None
            if not resp.status == 200:
                raise errors.ApiError("[%s]: %s" % (server, resp.text))
            default_name = path.split("/")[-1]
            config_data = {'name': default_name}
            try:
                metadata = await resp.json()
                config_data = json.loads(metadata['config_key__'])
                if config_data['name'] == '':
                    # use the path name unless there is a specifically configured value
                    config_data['name'] = path.split("/")[-1]
            except (KeyError, ValueError):
                # missing or corrupt configuration data
                pass

        # now get the stream info data
        url = "{server}/stream/list".format(server=server)
        params = {"path": path, "extended": '1'}
        async with session.get(url, params=params) as resp:

            info = (await resp.json())[0]
            my_stream = stream.from_nilmdb_metadata(config_data, info[1])
            my_info = StreamInfo(start=info[2],
                                 end=info[3],
                                 rows=info[4],
                                 bytes=-1,  # signal that this field is invalid
                                 total_time=info[5])
            return my_stream, my_info


async def _send_nilmdb_data(url, params, generator, dtype, session):
    bstart = params['start']
    async for data in generator:
        params['start'] = bstart
        np_data = np.frombuffer(data, dtype)
        bend = int(np_data['timestamp'][-1] + 1)
        params['end'] = bend

        async with session.put(url,
                               params=params,
                               data=data) as dest_response:
            if dest_response.status != 200:
                msg = await dest_response.text()
                raise errors.ApiError("Error writing to destination: %s" % msg)
        bstart = bend


async def _validate_nilmdb_url(url):
    async with aiohttp.ClientSession() as session:
        async with session.get(url) as resp:
            body = await resp.text()
            if 'NilmDB' not in body:
                raise errors.ApiError("[%s] is not a Nilmdb server" % url)


# save retrieved annotations so we don't make extra network requests
# there is no interval retrieval mechanism for NilmDB annotations
cached_nilmdb_annotations: Dict[str, List[Annotation]] = {}


async def _get_nilmdb_annotations(server: str, path: str, istart: int, iend: int) -> List[Annotation]:
    url = "{server}/stream/get_metadata".format(server=server)
    params = {"path": path, "key": '__annotations'}
    global cached_nilmdb_annotations
    if path in cached_nilmdb_annotations:
        annotations = cached_nilmdb_annotations[path]
        return [a for a in annotations if istart <= a.start <= iend]

    annotations = []
    async with aiohttp.ClientSession() as session:
        async with session.get(url, params=params) as resp:
            if resp.status == 404:
                return []
            if not resp.status == 200:
                raise errors.ApiError("[%s]: %s" % (server, resp.text))
            try:
                metadata = await resp.json()
                if metadata['__annotations'] is not None:
                    annotation_data = json.loads(metadata['__annotations'])
                    for item in annotation_data:
                        item['stream_id'] = None
                        annotations.append(annotation_from_json(item))
            except (KeyError, ValueError):
                # missing or corrupt annotation data
                pass
    cached_nilmdb_annotations[path] = annotations
    # only return annotations within the requested interval
    return [a for a in annotations if istart <= a.start <= iend]


async def _create_nilmdb_annotations(server: str, path: str, annotations: List[Annotation]) -> None:
    # create the JSON data
    annotation_json = [a.to_json() for a in annotations]
    data = {"__annotations": json.dumps(annotation_json)}
    url = "{server}/stream/update_metadata".format(server=server)
    data = {"path": path,
            "data": json.dumps(data)}
    async with aiohttp.ClientSession() as session:
        async with session.post(url, data=data) as resp:
            if not resp.status == 200:
                raise click.ClickException("cannot create metadata for [%s] on [%s]" % (path, server))



from sqlalchemy.orm import relationship
from sqlalchemy import (Column, Integer, String,
                        Boolean, Enum, ForeignKey)
from sqlalchemy.dialects.postgresql import BIGINT
from typing import List, Dict, TYPE_CHECKING
import configparser
import enum
import re
from operator import attrgetter

from joule.models.meta import Base
from joule.errors import ConfigurationError
from joule.models.data_store.data_store import StreamInfo
from joule.models import element, annotation

if TYPE_CHECKING:
    from joule.models import (Folder)  # pragma: no cover


class Stream(Base):
    """
    Attributes:
        name (str): stream name
        description (str): stream description
        datatype (Stream.DATATYPE): data representation on disk see :ref:`sec-streams`
        decimate (bool): whether to store decimated data for stream visualization
        folder (joule.Folder): parent Folder
        elements (List[joule.Element]): array of stream elements
        keep_us (int): microseconds of data to keep (KEEP_NONE=0, KEEP_ALL=-1).

    """
    __tablename__ = 'stream'
    __table_args__ = {"schema": "metadata"}

    id: int = Column(Integer, primary_key=True)
    name: str = Column(String, nullable=False)

    class DATATYPE(enum.Enum):
        FLOAT64 = enum.auto()
        FLOAT32 = enum.auto()
        INT64 = enum.auto()
        INT32 = enum.auto()
        INT16 = enum.auto()
        INT8 = enum.auto()
        UINT64 = enum.auto()
        UINT32 = enum.auto()
        UINT16 = enum.auto()
        UINT8 = enum.auto()

    datatype: DATATYPE = Column(Enum(DATATYPE), nullable=False)
    decimate: bool = Column(Boolean, default=True)

    # do not allow property changes if any of the following are true
    is_source: bool = Column(Boolean, default=False)
    is_destination: bool = Column(Boolean, default=False)
    is_configured: bool = Column(Boolean, default=False)

    KEEP_ALL = -1
    KEEP_NONE = 0
    keep_us: int = Column(BIGINT, default=KEEP_ALL)

    description: str = Column(String)
    folder_id: int = Column(Integer, ForeignKey('metadata.folder.id'))
    folder: "Folder" = relationship("Folder", back_populates="streams")
    elements: List[element.Element] = relationship("Element",
                                                   cascade="all, delete-orphan",
                                                   back_populates="stream")
    annotations: List[annotation.Annotation] = relationship("Annotation",
                                                            cascade="all, delete-orphan",
                                                            back_populates="stream")

    def merge_configs(self, other: 'Stream') -> None:
        # replace configurable attributes with other's values
        self.keep_us = other.keep_us
        self.decimate = other.decimate
        self.description = other.description
        self.is_configured = other.is_configured
        self.is_destination = other.is_destination
        self.is_source = other.is_source
        self.elements = other.elements

    def update_attributes(self, attrs: Dict) -> None:
        if 'name' in attrs:
            self.name = validate_name(attrs['name'])
        if 'description' in attrs:
            self.description = attrs['description']
        if 'elements' in attrs:
            element_configs = attrs['elements']
            # make sure the number of configs is correct
            if len(element_configs) != len(self.elements):
                raise ConfigurationError("incorrect number of elements")
            for e in self.elements:
                e.update_attributes(element_configs[e.index])

    def __str__(self):
        return "Stream [{name}]".format(name=self.name)

    def __repr__(self):
        return "<Stream(id=%r, name='%s', datatype=%r)>" % (
            self.id, self.name, self.datatype)

    @property
    def locked(self):
        """
        bool: true if the stream has a configuration file or is an active
         part of the data pipeline. Attributes of locked streams cannot be changed.
        """
        return self.is_configured or self.is_destination or self.is_source

    @property
    def active(self):
        """
        bool: true if the stream is part of the data pipeline
        """
        return self.is_source or self.is_destination

    @property
    def layout(self):
        """
        str: formatted string specifying the datatype and number of elements
        """
        return "%s_%d" % (self.datatype.name.lower(), len(self.elements))

    @property
    def decimated_layout(self):
        # decimations are floats (min,mean,max) tuples
        return "float32_%d" % (len(self.elements) * 3)

    @property
    def data_width(self) -> int:
        return len(self.elements) + 1

    @property
    def is_remote(self) -> bool:
        """
        bool: true if the stream resides on a remote system
        """
        try:
            return self._remote_node is not None
        except AttributeError:
            return False

    @property
    def remote_node(self) -> str:
        """
        str: URL of remote host, blank if the stream is local
        """
        try:
            return self._remote_node
        except AttributeError:
            return ''

    @property
    def remote_path(self) -> str:
        """
        str: path on remote host, blank if the stream is local
        """
        try:
            return self._remote_path
        except AttributeError:
            return ''

    def set_remote(self, node: str, path: str):
        """
        Associate the stream with a remote system
        Args:
            url: remote URL
            path: stream path on remote system
        """
        self._remote_node = node
        self._remote_path = path

    def to_json(self, info: Dict[int, StreamInfo] = None) -> Dict:
        """
          Args:
            info: optional content added to ``data_info`` field

        Returns: Dictionary of Stream attributes

        """

        if info is not None and self.id in info:
            data_info = info[self.id].to_json()
        else:
            data_info = None

        return {
            'id': self.id,
            'name': self.name,
            'description': self.description,
            'datatype': self.datatype.name.lower(),
            'layout': self.layout,
            'keep_us': self.keep_us,
            'is_configured': self.is_configured,
            'is_source': self.is_source,
            'is_destination': self.is_destination,
            'locked': self.locked,  # meta attribute
            'active': self.active,  # meta attribute
            'decimate': self.decimate,
            'elements': [e.to_json() for e in sorted(self.elements, key=attrgetter('index'))],
            'data_info': data_info
        }

    def to_nilmdb_metadata(self) -> Dict:
        return {
            'name': self.name,
            'name_abbrev': '',
            'delete_locked': False,
            'streams': [e.to_nilmdb_metadata() for e in self.elements]
        }


def from_json(data: Dict) -> Stream:
    """
    Construct a Stream from a dictionary of attributes produced by
    :meth:`Stream.to_json`

    Args:
        data: attribute dictionary

    Returns: Stream

    """
    elements = []
    index = 0
    for item in data["elements"]:
        item["index"] = index
        elements.append(element.from_json(item))
        index += 1

    return Stream(id=data["id"],
                  name=validate_name(data["name"]),
                  description=data["description"],
                  datatype=validate_datatype(data["datatype"].upper()),
                  keep_us=data["keep_us"],
                  decimate=data["decimate"],
                  is_configured=data["is_configured"],
                  is_source=data["is_source"],
                  is_destination=data["is_destination"],
                  elements=elements)


def from_config(config: configparser.ConfigParser) -> Stream:
    """
    Construct a Stream from a configuration file

    Args:
        config: parsed *.conf file

    Returns: Stream

    Raises: ConfigurationError

    """
    try:
        main_configs: configparser.ConfigParser = config["Main"]
    except KeyError as e:
        raise ConfigurationError("Missing section [%s]" % e) from e
    try:
        datatype = validate_datatype(main_configs["datatype"])
        keep_us = validate_keep(main_configs.get("keep", fallback="all"))
        decimate = main_configs.getboolean("decimate", fallback=True)
        name = validate_name(main_configs["name"])
        description = main_configs.get("description", fallback="")
        stream = Stream(name=name, description=description,
                        datatype=datatype, keep_us=keep_us, decimate=decimate)
    except KeyError as e:
        raise ConfigurationError("[Main] missing %s" % e.args[0]) from e
    # now try to load the elements
    element_configs = filter(lambda sec: re.match(r"Element\d", sec),
                             config.sections())
    index = 0
    for name in element_configs:
        try:
            e = element.from_config(config[name])
            e.index = index
            index += 1
            stream.elements.append(e)
        except ConfigurationError as e:
            raise ConfigurationError("element <%s> %s" % (name, e)) from e
    # make sure we have at least one element
    if len(stream.elements) == 0:
        raise ConfigurationError(
            "missing element configurations, must have at least one")
    return stream


def from_nilmdb_metadata(config_data: Dict, layout: str) -> Stream:
    datatype = validate_datatype(layout.split('_')[0])
    nelem = int(layout.split('_')[1])
    if 'streams' in config_data:
        elements = config_data['streams']
    else:
        elements = [{'column': i, 'name': "Element%d" % (i+1), 'units': None,
                     'scale_factor': 1.0, 'offset': 0.0, 'plottable': True,
                     'discrete': False, 'default_min': None, 'default_max': None} for
                    i in range(nelem)]
    stream = Stream(name=config_data['name'], description='', datatype=datatype,
                    keep_us=Stream.KEEP_ALL, decimate=True)
    for metadata in elements:
        elem = element.from_nilmdb_metadata(metadata)
        stream.elements.append(elem)
    return stream


def validate_name(name: str) -> str:
    if name is None or len(name) == 0:
        raise ConfigurationError("missing name")
    if '/' in name:
        raise ConfigurationError("invalid name, '\\' not allowed")
    return name


def validate_datatype(datatype: str) -> Stream.DATATYPE:
    try:
        return Stream.DATATYPE[datatype.upper()]
    except KeyError as e:
        valid_types = ", ".join([m.name.lower() for m in Stream.DATATYPE])
        raise ConfigurationError("invalid datatype [%s], choose from [%s]" %
                                 (datatype, valid_types)) from e


def validate_keep(keep: str) -> int:
    if keep.lower() == "none":
        return Stream.KEEP_NONE
    if keep.lower() == "all":
        return Stream.KEEP_ALL
    match = re.fullmatch(r'^(\d+)([hdwmy])$', keep)
    if match is None:
        raise ConfigurationError("invalid [Stream] keep"
                                 "use format #unit (eg 1w), none or all")

    units = {
        'h': 60 * 60 * 1e6,  # hours
        'd': 24 * 60 * 60 * 1e6,  # days
        'w': 7 * 24 * 60 * 60 * 1e6,  # weeks
        'm': 4 * 7 * 24 * 60 * 60 * 1e6,  # months
        'y': 365 * 24 * 60 * 60 * 1e6  # years
    }
    unit = match.group(2)
    time = int(match.group(1))
    if time <= 0:
        raise ConfigurationError("invalid [Stream] keep"
                                 "use format #unit (eg 1w), none or all")
    return int(time * units[unit])



import click
import asyncio
from treelib import Tree

from joule import errors
from joule.api.stream import Stream
from joule.api.folder import (Folder, folder_root)
from joule.cli.config import pass_config
from joule.api import BaseNode

@click.command(name="list")
@click.option("--layout", "-l", is_flag=True, help="include stream layout")
@click.option("--status", "-s", is_flag=True, help="include stream status")
@click.option("--id", "-i", is_flag=True, help="show ID's")
@pass_config
def cli_list(config, layout, status, id):
    loop = asyncio.get_event_loop()
    try:
        loop.run_until_complete(
            _run(config.node, layout, status, id))
    except errors.ApiError as e:
        raise click.ClickException(str(e)) from e
    finally:
        loop.run_until_complete(
            config.close_node())
        loop.close()


async def _run(node: BaseNode, layout: bool, status: bool, showid: bool):
    tree = Tree()
    root = await node.folder_root()
    root.name = ""  # omit root name
    _process_folder(tree, root, None, layout, status, showid)
    if status:
        click.echo("Legend: " + click.style("\u25CF ", fg="green") + "active  " +
                   click.style("\u25CF ", fg="cyan") + "configured")
    click.echo(tree.show())


def _process_folder(tree: Tree, folder: Folder, parent_id,
                    layout: bool, status: bool, showid: bool):
    tag = click.style(folder.name, bold=True)
    if showid:
        tag += " (%d)" % folder.id
    identifier = "f%d" % folder.id
    tree.create_node(tag, identifier, parent_id)
    for stream in folder.streams:
        _process_stream(tree, stream, identifier, layout, status, showid)
    for child in folder.children:
        _process_folder(tree, child, identifier, layout, status, showid)


def _process_stream(tree: Tree, stream: Stream, parent_id,
                    layout: bool, status: bool, showid: bool):
    tag = stream.name
    if showid:
        tag += " (%d)" % stream.id
    if layout:
        tag += " (%s)" % stream.layout
    if status:
        if stream.active:
            tag = click.style("\u25CF ", fg="green") + tag
        elif stream.locked:
            tag = click.style("\u25CF ", fg="cyan") + tag

    identifier = "s%d" % stream.id
    tree.create_node(tag, identifier, parent_id)



import numpy as np
import textwrap
from ..fir_filter_module import FIRFilterModule

ARGS_DESC = """
---
:name:
  Mean Filter
:author:
  John Donnal
:license:
  Open
:url:
  http://git.wattsworth.net/wattsworth/joule.git
:description:
  adjustable moving average
:usage:
    Apply a moving average to the input stream with an
    adjustable window size
    
    | Arguments | Description                     |
    |-----------|---------------------------------|
    |``window`` | samples to average, must be odd |


:inputs:
  input
  :  **<any type>** with N elements

:outputs:
  output
  :  **float32** with N elements

:stream_configs:
  #input#
     [Main]
     name = Raw Data
     path = /path/to/input
     datatype = int32
     keep = 1w

     [Element1]
     name = Element 1

     [Element2]
     name = Element 2

     #additional elements...

  #output#
     [Main]
     name = Filtered Data
     path = /path/to/output
     datatype = float32
     keep = 1w

     #same number of elements as input
     [Element1]
     name = Element 1

     [Element2]
     name = Element 2

     #additional elements...

:module_config:
    [Main]
    name = Mean Filter
    exec_cmd = joule-mean-filter

    [Arguments]
    # must be odd
    window = 11 

    [Inputs]
    input = /path/to/input

    [Outputs]
    output = /path/to/output
---
"""


class MeanFilter(FIRFilterModule):
    """Compute the moving average of the input"""
    
    def custom_args(self, parser):  # pragma: no cover
        grp = parser.add_argument_group("module",
                                        "module specific arguments")
        
        grp.add_argument("--window", type=int, required=True,
                         help="window length (odd)")

        parser.description = textwrap.dedent(ARGS_DESC)

    def make_filter(self, parsed_args):
        window = parsed_args.window
        return np.ones((window,))/window

                
def main():  # pragma: no cover
    r = MeanFilter()
    r.start()

    
if __name__ == "__main__":
    main()



import asyncio
import logging

from . import base_module
from joule import errors


class FilterModule(base_module.BaseModule):

    async def setup(self, parsed_args, app, inputs, outputs):
        """
        Configure the module, executes before :meth:`run`

        Args:
            parsed_args:
            app:
            inputs:
            outputs:

        Returns:

        """
        pass

    async def run(self, parsed_args, inputs, outputs):
        """
        This method must be implemented. It should run in a loop, if it returns the module stops.

        Args:
            parsed_args: parsed command line arguments, configure with :meth:`joule.BaseModule.custom_args`
            inputs: pipe connections to input streams indexed by name (specified in the module configuration file).
            outputs: pipe connections to output streams indexed by name (specified in the module configuration file).

        .. code-block:: python

            class ModuleDemo(FilterModule):

                def run(self, parsed_args, inputs, outputs):
                    raw = inputs["raw"]
                    filtered = outputs["filtered"]
                    # this filter just passes the input through to the output
                     while(not self.stop_requested):
                        data = await raw.read()
                        await filtered.write(data)
                        raw.consume(len(data))

                #... other module code
        """
        assert False, "implement in child class"  # pragma: no cover

    async def run_as_task(self, parsed_args, app, loop) -> asyncio.Task:
        try:
            (pipes_in, pipes_out) = await self._build_pipes(parsed_args)
        except errors.ApiError as e:
            logging.error(str(e))
            return loop.create_task(asyncio.sleep(0))
        await self.setup(parsed_args, app, pipes_in, pipes_out)
        return loop.create_task(
            self.run(parsed_args, pipes_in, pipes_out))



import click
import asyncio

from joule.cli.config import pass_config
from joule import errors


@click.command(name="list")
@pass_config
def cli_list(config):
    loop = asyncio.get_event_loop()
    try:
        loop.run_until_complete(
            _run(config.node))
    except errors.ApiError as e:
        raise click.ClickException(str(e)) from e
    finally:
        loop.run_until_complete(
            config.close_node())
        loop.close()


async def _run(node):
    followers = await node.follower_list()
    # display follower information
    if len(followers) > 0:
        click.echo("This node can control:")
        for follower in followers:
            click.echo("\t%s" % follower.name)
    else:
        click.echo("This node cannot control any other nodes")



import click
from .list import cli_list
from .info import cli_info
from .logs import cli_logs


@click.group(name="module")
def module():
    pass  # pragma: no cover


module.add_command(cli_list)
module.add_command(cli_info)
module.add_command(cli_logs)



import click
from .list import cli_list
from .info import cli_info


@click.group(name="proxy")
def proxies():
    pass  # pragma: no cover


proxies.add_command(cli_list)
proxies.add_command(cli_info)



from aiohttp import web

from joule.models.supervisor import Supervisor


async def index(request):
    supervisor: Supervisor = request.app["supervisor"]
    resp = []
    if 'statistics' in request.query and request.query['statistics'] != '0':
        get_stats = True
    else:
        get_stats = False

    for worker in supervisor.workers:
        worker_info = {
            "id": worker.uuid,
            "name": worker.name,
            "description": worker.description,
            "is_app": worker.is_app,
            "inputs": {},
            "outputs": {}}
        if get_stats:
            worker_info['statistics'] = \
                (await worker.statistics()).to_json()

        for c in worker.input_connections:
            worker_info['inputs'][c.name] = c.location
        for c in worker.output_connections:
            worker_info['outputs'][c.name] = c.location
        resp.append(worker_info)
    return web.json_response(data=resp)


async def info(request):
    supervisor: Supervisor = request.app["supervisor"]
    if 'name' in request.query:
        worker = [w for w in supervisor.workers if w.name == request.query['name']]
    elif 'id' in request.query:
        worker = [w for w in supervisor.workers if w.uuid == int(request.query['id'])]
    else:
        return web.Response(text="specify a name or id", status=400)
    if len(worker) == 0:
        return web.Response(text="module does not exist", status=404)
    worker = worker[0]

    data = {
        "id": worker.uuid,
        "name": worker.name,
        "description": worker.description,
        "is_app": worker.is_app,
        "inputs": {},
        "outputs": {},
        "statistics": (await worker.statistics()).to_json()}
    for c in worker.input_connections:
        data['inputs'][c.name] = c.location
    for c in worker.output_connections:
        data['outputs'][c.name] = c.location
    return web.json_response(data)


async def logs(request):
    supervisor: Supervisor = request.app["supervisor"]
    if 'name' in request.query:
        worker = [w for w in supervisor.workers if w.name == request.query['name']]
    elif 'id' in request.query:
        worker = [w for w in supervisor.workers if w.uuid == int(request.query['id'])]
    else:
        return web.Response(text="specify a name", status=400)
    if len(worker) == 0:
        return web.Response(text="module does not exist", status=404)
    worker = worker[0]
    return web.json_response(worker.logs)



import click
from .info import node_info
from .list import node_list
from .add import node_add
from .delete import node_delete
from .default import node_default


@click.group(name="node")
def node():
    pass  # pragma: no cover


node.add_command(node_info)
node.add_command(node_list)
node.add_command(node_add)
node.add_command(node_delete)
node.add_command(node_default)



import enum
from typing import TYPE_CHECKING, List
import numpy as np
import logging

from joule.models.pipes.errors import PipeError

if TYPE_CHECKING:  # pragma: no cover
    from joule.models import (Module, Stream)

log = logging.getLogger('joule')


class Pipe:
    """
    This encapsulates streams and connects to modules.
    Some more infos2

    .. note::

       There are many different kinds of pipes

    """

    class DIRECTION(enum.Enum):
        INPUT = enum.auto()
        OUTPUT = enum.auto()
        TWOWAY = enum.auto()

    def __init__(self, name=None, direction=None, module=None, stream=None, layout=None):
        """

        :param name: optional name for the pipe, useful for debugging
        :param direction: data flow
        :type direction: DIRECTION
        :param module: what??
        :param stream: optional stream
        :param layout: the data layout
        """
        self.name: str = name
        self.direction: Pipe.DIRECTION = direction
        self.module: 'Module' = module
        self.stream: 'Stream' = stream
        self._layout = layout
        self.closed = False
        self.decimation_level = 1
        self.subscribers: List['Pipe'] = []
        self._failed = False

    def __repr__(self):
        return '<joule.models.Pipe name=%r direction=%r layout=%r>' % (
            self.name, self.direction.name, self.layout)

    def enable_cache(self, lines: int):
        """
        Turn on caching for pipe writes. Data is only transmitted once the cache is full.
        This improves system performance especially if :meth:`write` is called
        rapidly with short arrays. Once enabled, caching cannot be disabled.

        Args:
            lines: cache size

        """
        if self.direction == Pipe.DIRECTION.INPUT:
            raise PipeError("cannot control cache on input pipes")
        raise PipeError("abstract method must be implemented by child")

    async def flush_cache(self):
        """
        Force a pipe flush even if the cache is not full. Raises an error if caching is not
        enabled.

        """
        if self.direction == Pipe.DIRECTION.INPUT:
            raise PipeError("cannot control cache on input pipes")
        raise PipeError("abstract method must be implemented by child")

    async def read(self, flatten=False) -> np.ndarray:
        """
        Read stream data. By default this method returns a structured
        array with ``timestamp`` and ``data`` fields. This method is a coroutine.

        Args:
            flatten: return an unstructured array (flat 2D matrix) with timestamps in the first column

        Returns:
            numpy.ndarray

        >>> data = await pipe.read()
        [1, 2, 3]
        >>> data = await pipe.read(flatten=True)
        # the same data is returned again
        [1,2,3]
        >>> pipe.consume(len(data))
        # next call to read will return only new data
        """
        if self.direction == Pipe.DIRECTION.OUTPUT:
            raise PipeError("cannot read from an output pipe")

        raise PipeError("abstract method must be implemented by child")

    def reread_last(self):
        """
        The next read will return only unconsumed data from the previous read
        and no new data from the source. The end_of_interval flag is maintained.

        """
        if self.direction == Pipe.DIRECTION.OUTPUT:
            raise PipeError("cannot read from an output pipe")

        raise PipeError("Not Implemented")

    async def read_all(self, flatten=False, maxrows=1e5, error_on_overflow=False) -> np.ndarray:
        """
                Read stream data. By default this method returns a structured
                array with ``timestamp`` and ``data`` fields. The pipe is automatically closed.
                This method is a coroutine.

                Args:
                    flatten: return an unstructured array (flat 2D matrix) with timestamps in the first column
                    maxrows: the maximum number of rows to read from the pipe
                    error_on_overflow: raise a PipeError exception if pipe is not empty after reading maxrows

                Returns:
                    numpy.ndarray

                >>> data = await pipe.read_all(flatten=True)
                [1, 2, 3]
        """
        if self.direction == Pipe.DIRECTION.OUTPUT:
            raise PipeError("cannot read from an output pipe")

        data = None
        while True:
            try:
                new_data = await self.read(flatten)
                self.consume(len(new_data))
            except PipeError:
                break
            if data is None:
                data = new_data
                if len(data) > maxrows:
                    await self.close()
                    if error_on_overflow:
                        raise PipeError("More than [%d] rows, increase maxrows or disable error_on_overflow" % maxrows)
                    return data[:maxrows]
            else:
                if len(data) + len(new_data) > maxrows:
                    await self.close()
                    if error_on_overflow:
                        raise PipeError("More than [%d] rows, increase maxrows or disable error_on_overflow" % maxrows)
                    remaining_rows = maxrows - len(data)
                    if flatten:
                        data = np.vstack((data, new_data[:remaining_rows]))
                    else:
                        data = np.hstack((data, new_data[:remaining_rows]))
                    break
                if flatten:
                    data = np.vstack((data, new_data))
                else:
                    data = np.hstack((data, new_data))
        if data is None:
            raise PipeError("No data in pipe")
        return data

    def consume(self, num_rows):
        """
        Flush data from the read buffer. The next call to :meth:`read` will
        return any unflushed data followed by new incoming data.

        Args:
            num_rows: number of rows to flush from the read buffer

        """

        if self.direction == Pipe.DIRECTION.OUTPUT:
            raise PipeError("cannot consume from an output pipe")
        raise PipeError("abstract method must be implemented by child")

    def fail(self):
        self._failed = True

    @property
    def end_of_interval(self):
        return False

    async def write(self, data):
        """
        Write timestamped data to the pipe. Timestamps must be monotonically increasing
        and should not overlap with existing stream data in the database. This method is a coroutine.

        Args:
            data (numpy.ndarray): May be a structured array with ``timestamp`` and ``data`` fields
            or an unstructured array with timestamps in the first column.

        >>> await pipe.write([[1000, 2, 3],[1001, 3, 4]])

        """
        if self.direction == Pipe.DIRECTION.INPUT:
            raise PipeError("cannot write to an input pipe")
        raise PipeError("abstract method must be implemented by child")

    async def close_interval(self):
        """
        Signal a break in the data stream. This should be used to indicate missing data.
        Data returned from :meth:`read` will be chunked by interval boundaries.

        """
        if self.direction == Pipe.DIRECTION.INPUT:
            raise PipeError("cannot write to an input pipe")
        raise PipeError("abstract method must be implemented by child")

    def close_interval_nowait(self):
        """
        Signal a break in the data stream. This will dumped cached data and should generally
        not be used. Instead use the coroutine :meth:`close_interval`.

        """
        pass  # pragma: no cover

    def subscribe(self, pipe):
        if self.direction == Pipe.DIRECTION.INPUT:
            raise PipeError("cannot subscribe to an input pipe")

        self.subscribers.append(pipe)

        def unsubscribe():
            i = self.subscribers.index(pipe)
            del self.subscribers[i]

        return unsubscribe

    async def close(self):
        """
        Close the pipe. This also closes any subscribers. If ``close_cb`` is defined
        it will be executed before the subscribers are closed.

        """
        # close the pipe, optionally implemented by the child
        pass  # pragma: no cover

    def change_layout(self, layout: str):
        raise PipeError("layout cannot be changed")

    @property
    def layout(self) -> str:
        if self._layout is not None:
            return self._layout
        elif self.stream is not None:
            return self.stream.layout
        else:
            raise ValueError("pipe has no layout")

    @property
    def width(self) -> int:
        shape = self.dtype['data'].shape
        if len(shape) == 0:
            return 1
        else:
            return shape[0]

    @property
    def dtype(self) -> np.dtype:
        return compute_dtype(self.layout)

    @property
    def decimated(self) -> bool:
        if self.decimation_level > 1:
            return True
        else:
            return False

    def _apply_dtype(self, data: np.ndarray) -> np.ndarray:
        """convert [data] to the pipe's [dtype]"""
        if data.ndim == 1:
            # already a structured array just verify its data type
            if data.dtype != self.dtype:
                raise PipeError("wrong dtype for 1D (structured) array" +
                                "[%s] != req type [%s]" % (data.dtype,
                                                           self.dtype))
            return data
        elif data.ndim == 2:
            # Convert to structured array
            sarray = np.zeros(data.shape[0], dtype=self.dtype)
            try:
                sarray['timestamp'] = data[:, 0]
                # Need the squeeze in case sarray['data'] is 1 dimensional
                sarray['data'] = np.squeeze(data[:, 1:])
                return sarray
            except (IndexError, ValueError):
                raise PipeError("wrong number of fields for this data type")
        else:
            raise PipeError("wrong number of dimensions in array")

    @staticmethod
    def _format_data(data, flatten):
        if flatten:
            return np.c_[data['timestamp'][:, None], data['data']]
        else:
            return data

    @staticmethod
    def _validate_data(data):
        if type(data) is not np.ndarray:
            raise PipeError("invalid data type must be a structured array or 2D matrix")
        # check for valid data type
        try:
            if (len(data) == 0) or len(data[0]) == 0:
                log.info("pipe write called with no data")
                return False
        except TypeError:
            raise PipeError("invalid data type must be a structured array or 2D matrix")
        return True


def interval_token(layout):
    nelem = int(layout.split('_')[1])
    token = np.zeros(1, dtype=compute_dtype(layout))
    token['timestamp'] = 0
    token['data'] = np.zeros(nelem)
    return token


def find_interval_token(raw: bytes, layout):
    token = interval_token(layout)[0]
    token_bytes = token.tobytes()
    index = raw.find(token_bytes)
    if index == -1:
        return None
    # this data *may* have an interval, need to check if the token
    # is aligned to a row
    data = np.frombuffer(raw, dtype=compute_dtype(layout))
    i = 0
    for row in data:
        if np.array_equal(row, token):
            return i * len(token_bytes), (i + 1) * len(token_bytes)
        i += 1

    # suprious match, not aligned to a row
    return None


def compute_dtype(layout: str) -> np.dtype:
    try:
        ltype = layout.split('_')[0]
        lcount = int(layout.split('_')[1])
        if ltype.startswith('int'):
            atype = '<i' + str(int(ltype[3:]) // 8)
        elif ltype.startswith('uint'):
            atype = '<u' + str(int(ltype[4:]) // 8)
        elif ltype.startswith('float'):
            atype = '<f' + str(int(ltype[5:]) // 8)
        else:
            raise ValueError()
        return np.dtype([('timestamp', '<i8'), ('data', atype, lcount)])
    except (IndexError, ValueError):
        raise ValueError("bad layout: [%s]" % layout)



from sqlalchemy.orm import Session
from typing import List

from joule.errors import ConfigurationError
from joule.models import Stream, Element, Follower
from joule.models import stream, folder


# basic format of stream is just the full_path
# /file/path/stream_name
# but streams may be configured inline
# /file/path/stream_name:datatype[e1, e2, e3, ...]
#  where e1, e2, e2, ... are the element names
#


def run(pipe_config: str, db: Session) -> Stream:
    # check for a remote stream config
    (pipe_config, node_name) = strip_remote_config(pipe_config)
    local = node_name is None
    # separate the configuration pieces
    (path, name, inline_config) = parse_pipe_config(pipe_config)
    name = stream.validate_name(name)
    # parse the inline configuration
    (datatype, element_names) = parse_inline_config(inline_config)
    # if the stream is local, check for it in the database
    if local:
        my_folder = folder.find(path, db, create=True)
        # check if the stream exists in the database
        existing_stream: Stream = db.query(Stream). \
            filter_by(folder=my_folder, name=name). \
            one_or_none()
        if existing_stream is not None:
            if len(inline_config) > 0:
                _validate_config_match(existing_stream, datatype, element_names)
            return existing_stream
    else:  # make sure the remote node is a follower
        if db.query(Follower).filter_by(name=node_name).one_or_none() is None:
            raise ConfigurationError("Remote node [%s] is not a follower" % node_name)

    # if the stream doesn't exist it or its remote, it *must* have inline configuration
    if len(inline_config) == 0:
        if local:
            msg = "add inline config or *.conf file for stream [%s]" % pipe_config
        else:
            msg = "remote streams must have inline config"
        raise ConfigurationError(msg)

    # build the stream from inline config
    my_stream = stream.Stream(name=name,
                              datatype=datatype)
    i = 0
    for e in element_names:
        my_stream.elements.append(Element(name=e, index=i))
        i += 1
    if local:
        my_folder.streams.append(my_stream)
        db.add(my_stream)
    else:
        my_stream.set_remote(node_name, path + '/' + my_stream.name)
    return my_stream


def strip_remote_config(pipe_config: str) -> (str, str):
    try:
        # check for the remote URL return stripped config and info
        if pipe_config[0] == '/':
            return pipe_config, None
        # this is a remote stream, separate out the node name
        pieces = pipe_config.split(' ')
        node = pieces[0]
        pipe_config = pieces[1]
    except (ValueError, IndexError):
        raise ConfigurationError("invalid pipe configuration [%s]" % pipe_config)
    return pipe_config, node


def parse_pipe_config(pipe_config: str) -> (str, str, str):
    # convert /path/name:datatype[e0,e1,e2] into (/path, name, datatype[e0,e1,e2])
    if ':' in pipe_config:
        full_path = pipe_config.split(':')[0]
        inline_config = pipe_config.split(':')[1]
    else:
        full_path = pipe_config
        inline_config = ""
    if '/' not in full_path:  # pragma: no cover
        # this is a double check, missing / will be caught earlier
        raise ConfigurationError("invalid path [%s]" % pipe_config)
    path_chunks = full_path.split('/')
    name = path_chunks.pop()
    if name == "":
        raise ConfigurationError("invalid stream name [%s]" % pipe_config)
    path = '/'.join(path_chunks)
    return path, name, inline_config


def parse_inline_config(inline_config: str) -> (Stream.DATATYPE, List[str]):
    if len(inline_config) == 0:
        return None, None
    try:
        config = inline_config.split('[')
        if len(config) != 2 or inline_config[-1] != ']':
            raise ConfigurationError("format is datatype[e1,e2,e3,...]")
        datatype = stream.validate_datatype(config[0].strip())
        element_names = [e.strip() for e in config[1].strip()[:-1].split(",")]
        return datatype, element_names
    except ConfigurationError as e:
        raise ConfigurationError("invalid inline configuration: %s" % inline_config) from e


def _validate_config_match(existing_stream: Stream,
                           datatype: Stream.DATATYPE,
                           elem_names: List[str]):
    # verify data_type match if specified
    if (datatype is not None and
            datatype != existing_stream.datatype):
        raise ConfigurationError("Stream [%s] exists with data_type %s" %
                                 (existing_stream.name,
                                  existing_stream.datatype))

    # verify elem_names
    existing_names = [e.name for e in existing_stream.elements]
    if len(existing_names) != len(elem_names):
        raise ConfigurationError("Stream exists with different elements")
    for name in elem_names:
        if name not in existing_names:
            raise ConfigurationError("Stream exists with different elements")
